############ colormap ###############

colormap_crc98_ct_group = {
    "Tumor cE01 (Stem/TA-like)": "#ff9999",
    "Tumor cE02 (Stem/TA-like/Immature Goblet)": "#ffa599",
    "Tumor cE03 (Stem/TA-like prolif)": "#ecf4dd",
    "Tumor cE04 (Enterocyte 1)": "#ffbd99",
    "Tumor cE05 (Enterocyte 2)": "#d8a7b1",
    "Tumor cE06 (Immature Goblet)": "#ffd699",
    "Tumor cE07 (Goblet/Enterocyte)": "#ffe299",
    "Tumor cE08 (Goblet)": "#ffee99",
    "Tumor cE09 (Best4)": "#fffa99",
    "Tumor cE10 (Tuft)": "#f6ff99",
    "Tumor cE11 (Enteroendocrine)": "#fe817d",
    "cB1 (B IGD+IgM+)": "#b2e87c",
    "cB2 (B GC-like)": "#aee87c",
    "cB3 (B CD40+ GC-like)": "#aae87c",
    "cE01 (Stem/TA-like)": "#a6e87c",
    "cE02 (Stem/TA-like/Immature Goblet)": "#a2e87c",
    "cE03 (Stem/TA-like prolif)": "#9de87c",
    "cE04 (Enterocyte 1)": "#99e87c",
    "cE05 (Enterocyte 2)": "#95e87c",
    "cE06 (Immature Goblet)": "#91e87c",
    "cE07 (Goblet/Enterocyte)": "#8de87c",
    "cE08 (Goblet)": "#89e87c",
    "cE09 (Best4)": "#85e87c",
    "cE10 (Tuft)": "#81e87c",
    "cE11 (Enteroendocrine)": "#7de87c",
    "cM01 (Monocyte)": "#00a087",
    "cM02 (Macrophage-like)": "#aed185",
    "cM03 (DC1)": "#7ce889",
    "cM04 (DC2)": "#7ce88d",
    "cM05 (DC2 C1Q+)": "#7ce891",
    "cM06 (DC IL22RA2)": "#7ce895",
    "cM07 (pDC)": "#7ce899",
    "cM08 (AS-DC)": "#7ce89d",
    "cM09 (mregDC)": "#7ce8a1",
    "cM10 (Granulocyte)": "#7ce8a5",
    "cMA01 (Mast)": "#7ce8aa",
    "cP1 (Plasma IgA)": "#7ce8ae",
    "cP2 (Plasma IgG)": "#efbb8a",
    "cP3 (Plasma IgG prolif)": "#af6c63",
    "cS01 (Endo arterial)": "#7ce8ba",
    "cS02 (Endo capillary)": "#7ce8be",
    "cS03 (Endo capillary)": "#c1e8ee",
    "cS04 (Endo)": "#92b4c8",
    "cS05 (Endo venous)": "#7ce8ca",
    "cS06 (Endo lymphatic)": "#7ce8cf",
    "cS07 (Endo capillary-like)": "#7ce8d3",
    "cS08 (Endo arterial-like)": "#7ce8d7",
    "cS09 (Endo)": "#7ce8db",
    "cS10 (Endo tip cells)": "#7ce8df",
    "cS11 (Endo proif)": "#7ce8e3",
    "cS12 (Endo)": "#7ce8e7",
    "cS13 (Endo venous-like)": "#7ce4e8",
    "cS14 (Endo)": "#9ec1d4",
    "cS15 (Pericyte)": "#7cdce8",
    "cS16 (Pericyte)": "#7cd7e8",
    "cS17 (Pericyte)": "#7cd3e8",
    "cS18 (Pericyte)": "#7ccfe8",
    "cS19 (Pericyte)": "#7ccbe8",
    "cS20 (Pericyte prolif)": "#7cc7e8",
    "cS21 (Fibro stem cell niche)": "#7cc3e8",
    "cS22 (Fibro stem cell niche)": "#7cbfe8",
    "cS23 (Fibro BMP-producing)": "#7cbbe8",
    "cS24 (Fibro BMP-producing)": "#7cb7e8",
    "cS25 (Fibro CCL8+)": "#7cb2e8",
    "cS26 (Myofibro)": "#7caee8",
    "cS27 (CXCL14+ CAF)": "#7caae8",
    "cS28 (GREM1+ CAF)": "#7ca6e8",
    "cS29 (MMP3+ CAF)": "#7ca2e8",
    "cS30 (CAF CCL8 Fibro-like)": "#7c9ee8",
    "cS31 (CAF stem niche Fibro-like)": "#7c9ae8",
    "cS32 (Smooth Muscle)": "#7c96e8",
    "cS33 (Schwann)": "#7c92e8",
    "cTNI01 (CD4+ IL7R+)": "#7c8ee8",
    "cTNI02 (CD4+ IL7R+SELL+)": "#7c89e8",
    "cTNI03 (CD4+ IL7R+HSP+)": "#7c85e8",
    "cTNI04 (CD4+ IL7R+CCL5+)": "#7c81e8",
    "cTNI05 (CD4+ IL17+)": "#7c7de8",
    "cTNI06 (CD4+ TFH)": "#807ce8",
    "cTNI07 (CD4+ CXCL13+)": "#847ce8",
    "cTNI08 (CD4+ Treg)": "#887ce8",
    "cTNI09 (CD4+ Treg prolif)": "#8c7ce8",
    "cTNI10 (CD8+ IL7R+)": "#907ce8",
    "cTNI11 (CD8+GZMK+)": "#947ce8",
    "cTNI12 (CD8+ IL7R+)": "#997ce8",
    "cTNI13 (CD8+ T IL17+)": "#9d7ce8",
    "cTNI14 (CD8+ CXCL13+)": "#a17ce8",
    "cTNI15 (CD8+ CXCL13+ HSP+)": "#a57ce8",
    "cTNI16 (CD8+ CXCL13+ prolif)": "#a97ce8",
    "cTNI17 (gd-like T)": "#ad7ce8",
    "cTNI18 (gd-like T PDCD1+)": "#b17ce8",
    "cTNI19 (gd-like T prolif)": "#b57ce8",
    "cTNI20 (PLZF+ T)": "#b97ce8",
    "cTNI21 (PLZF+ T prolif)": "#be7ce8",
    "cTNI22 (cTNI22)": "#c27ce8",
    "cTNI23 (NK CD16A+)": "#c67ce8",
    "cTNI24 (NK GZMK+)": "#ca7ce8",
    "cTNI25 (NK XCL1+)": "#ce7ce8",
    "cTNI26 (ILC3)": "#d27ce8",
}

colormap_coloncatlas_98 = {
    "p_ct_Tumor cE01 (Stem/TA-like)": "#ff9999",
    "p_ct_Tumor cE02 (Stem/TA-like/Immature Goblet)": "#ffa599",
    "p_ct_Tumor cE03 (Stem/TA-like prolif)": "#ecf4dd",
    "p_ct_Tumor cE04 (Enterocyte 1)": "#ffbd99",
    "p_ct_Tumor cE05 (Enterocyte 2)": "#d8a7b1",
    "p_ct_Tumor cE06 (Immature Goblet)": "#ffd699",
    "p_ct_Tumor cE07 (Goblet/Enterocyte)": "#ffe299",
    "p_ct_Tumor cE08 (Goblet)": "#ffee99",
    "p_ct_Tumor cE09 (Best4)": "#fffa99",
    "p_ct_Tumor cE10 (Tuft)": "#f6ff99",
    "p_ct_Tumor cE11 (Enteroendocrine)": "#fe817d",
    "p_ct_cB1 (B IGD+IgM+)": "#b2e87c",
    "p_ct_cB2 (B GC-like)": "#aee87c",
    "p_ct_cB3 (B CD40+ GC-like)": "#aae87c",
    "p_ct_cE01 (Stem/TA-like)": "#a6e87c",
    "p_ct_cE02 (Stem/TA-like/Immature Goblet)": "#a2e87c",
    "p_ct_cE03 (Stem/TA-like prolif)": "#9de87c",
    "p_ct_cE04 (Enterocyte 1)": "#99e87c",
    "p_ct_cE05 (Enterocyte 2)": "#95e87c",
    "p_ct_cE06 (Immature Goblet)": "#91e87c",
    "p_ct_cE07 (Goblet/Enterocyte)": "#8de87c",
    "p_ct_cE08 (Goblet)": "#89e87c",
    "p_ct_cE09 (Best4)": "#85e87c",
    "p_ct_cE10 (Tuft)": "#81e87c",
    "p_ct_cE11 (Enteroendocrine)": "#7de87c",
    "p_ct_cM01 (Monocyte)": "#00a087",
    "p_ct_cM02 (Macrophage-like)": "#aed185",
    "p_ct_cM03 (DC1)": "#7ce889",
    "p_ct_cM04 (DC2)": "#7ce88d",
    "p_ct_cM05 (DC2 C1Q+)": "#7ce891",
    "p_ct_cM06 (DC IL22RA2)": "#7ce895",
    "p_ct_cM07 (pDC)": "#7ce899",
    "p_ct_cM08 (AS-DC)": "#7ce89d",
    "p_ct_cM09 (mregDC)": "#7ce8a1",
    "p_ct_cM10 (Granulocyte)": "#7ce8a5",
    "p_ct_cMA01 (Mast)": "#7ce8aa",
    "p_ct_cP1 (Plasma IgA)": "#7ce8ae",
    "p_ct_cP2 (Plasma IgG)": "#efbb8a",
    "p_ct_cP3 (Plasma IgG prolif)": "#af6c63",
    "p_ct_cS01 (Endo arterial)": "#7ce8ba",
    "p_ct_cS02 (Endo capillary)": "#7ce8be",
    "p_ct_cS03 (Endo capillary)": "#c1e8ee",
    "p_ct_cS04 (Endo)": "#92b4c8",
    "p_ct_cS05 (Endo venous)": "#7ce8ca",
    "p_ct_cS06 (Endo lymphatic)": "#7ce8cf",
    "p_ct_cS07 (Endo capillary-like)": "#7ce8d3",
    "p_ct_cS08 (Endo arterial-like)": "#7ce8d7",
    "p_ct_cS09 (Endo)": "#7ce8db",
    "p_ct_cS10 (Endo tip cells)": "#7ce8df",
    "p_ct_cS11 (Endo proif)": "#7ce8e3",
    "p_ct_cS12 (Endo)": "#7ce8e7",
    "p_ct_cS13 (Endo venous-like)": "#7ce4e8",
    "p_ct_cS14 (Endo)": "#9ec1d4",
    "p_ct_cS15 (Pericyte)": "#7cdce8",
    "p_ct_cS16 (Pericyte)": "#7cd7e8",
    "p_ct_cS17 (Pericyte)": "#7cd3e8",
    "p_ct_cS18 (Pericyte)": "#7ccfe8",
    "p_ct_cS19 (Pericyte)": "#7ccbe8",
    "p_ct_cS20 (Pericyte prolif)": "#7cc7e8",
    "p_ct_cS21 (Fibro stem cell niche)": "#7cc3e8",
    "p_ct_cS22 (Fibro stem cell niche)": "#7cbfe8",
    "p_ct_cS23 (Fibro BMP-producing)": "#7cbbe8",
    "p_ct_cS24 (Fibro BMP-producing)": "#7cb7e8",
    "p_ct_cS25 (Fibro CCL8+)": "#7cb2e8",
    "p_ct_cS26 (Myofibro)": "#7caee8",
    "p_ct_cS27 (CXCL14+ CAF)": "#7caae8",
    "p_ct_cS28 (GREM1+ CAF)": "#7ca6e8",
    "p_ct_cS29 (MMP3+ CAF)": "#7ca2e8",
    "p_ct_cS30 (CAF CCL8 Fibro-like)": "#7c9ee8",
    "p_ct_cS31 (CAF stem niche Fibro-like)": "#7c9ae8",
    "p_ct_cS32 (Smooth Muscle)": "#7c96e8",
    "p_ct_cS33 (Schwann)": "#7c92e8",
    "p_ct_cTNI01 (CD4+ IL7R+)": "#7c8ee8",
    "p_ct_cTNI02 (CD4+ IL7R+SELL+)": "#7c89e8",
    "p_ct_cTNI03 (CD4+ IL7R+HSP+)": "#7c85e8",
    "p_ct_cTNI04 (CD4+ IL7R+CCL5+)": "#7c81e8",
    "p_ct_cTNI05 (CD4+ IL17+)": "#7c7de8",
    "p_ct_cTNI06 (CD4+ TFH)": "#807ce8",
    "p_ct_cTNI07 (CD4+ CXCL13+)": "#847ce8",
    "p_ct_cTNI08 (CD4+ Treg)": "#887ce8",
    "p_ct_cTNI09 (CD4+ Treg prolif)": "#8c7ce8",
    "p_ct_cTNI10 (CD8+ IL7R+)": "#907ce8",
    "p_ct_cTNI11 (CD8+GZMK+)": "#947ce8",
    "p_ct_cTNI12 (CD8+ IL7R+)": "#997ce8",
    "p_ct_cTNI13 (CD8+ T IL17+)": "#9d7ce8",
    "p_ct_cTNI14 (CD8+ CXCL13+)": "#a17ce8",
    "p_ct_cTNI15 (CD8+ CXCL13+ HSP+)": "#a57ce8",
    "p_ct_cTNI16 (CD8+ CXCL13+ prolif)": "#a97ce8",
    "p_ct_cTNI17 (gd-like T)": "#ad7ce8",
    "p_ct_cTNI18 (gd-like T PDCD1+)": "#b17ce8",
    "p_ct_cTNI19 (gd-like T prolif)": "#b57ce8",
    "p_ct_cTNI20 (PLZF+ T)": "#b97ce8",
    "p_ct_cTNI21 (PLZF+ T prolif)": "#be7ce8",
    "p_ct_cTNI22 (cTNI22)": "#c27ce8",
    "p_ct_cTNI23 (NK CD16A+)": "#c67ce8",
    "p_ct_cTNI24 (NK GZMK+)": "#ca7ce8",
    "p_ct_cTNI25 (NK XCL1+)": "#ce7ce8",
    "p_ct_cTNI26 (ILC3)": "#d27ce8",
    "ambiguous": "#d3d3d3",
    "filtered": "#848884",
}

colormap_coloncatlas_98_light = {
    "p_ct_Tumor cE01 (Stem/TA-like)": "#ffe5e5",
    "p_ct_Tumor cE02 (Stem/TA-like/Immature Goblet)": "#ffe8e5",
    "p_ct_Tumor cE03 (Stem/TA-like prolif)": "#fafcf6",
    "p_ct_Tumor cE04 (Enterocyte 1)": "#ffeee5",
    "p_ct_Tumor cE05 (Enterocyte 2)": "#f5e9eb",
    "p_ct_Tumor cE06 (Immature Goblet)": "#fff4e5",
    "p_ct_Tumor cE07 (Goblet/Enterocyte)": "#fff7e5",
    "p_ct_Tumor cE08 (Goblet)": "#fffae5",
    "p_ct_Tumor cE09 (Best4)": "#fffde5",
    "p_ct_Tumor cE10 (Tuft)": "#fcffe5",
    "p_ct_Tumor cE11 (Enteroendocrine)": "#fedfde",
    "p_ct_cB1 (B IGD+IgM+)": "#ebf9de",
    "p_ct_cB2 (B GC-like)": "#eaf9de",
    "p_ct_cB3 (B CD40+ GC-like)": "#e9f9de",
    "p_ct_cE01 (Stem/TA-like)": "#e8f9de",
    "p_ct_cE02 (Stem/TA-like/Immature Goblet)": "#e7f9de",
    "p_ct_cE03 (Stem/TA-like prolif)": "#e6f9de",
    "p_ct_cE04 (Enterocyte 1)": "#e5f9de",
    "p_ct_cE05 (Enterocyte 2)": "#e4f9de",
    "p_ct_cE06 (Immature Goblet)": "#e3f9de",
    "p_ct_cE07 (Goblet/Enterocyte)": "#e2f9de",
    "p_ct_cE08 (Goblet)": "#e1f9de",
    "p_ct_cE09 (Best4)": "#e0f9de",
    "p_ct_cE10 (Tuft)": "#dff9de",
    "p_ct_cE11 (Enteroendocrine)": "#def9de",
    "p_ct_cM01 (Monocyte)": "#bfe7e1",
    "p_ct_cM02 (Macrophage-like)": "#eaf3e0",
    "p_ct_cM03 (DC1)": "#def9e1",
    "p_ct_cM04 (DC2)": "#def9e2",
    "p_ct_cM05 (DC2 C1Q+)": "#def9e3",
    "p_ct_cM06 (DC IL22RA2)": "#def9e4",
    "p_ct_cM07 (pDC)": "#def9e5",
    "p_ct_cM08 (AS-DC)": "#def9e6",
    "p_ct_cM09 (mregDC)": "#def9e7",
    "p_ct_cM10 (Granulocyte)": "#def9e8",
    "p_ct_cMA01 (Mast)": "#def9e9",
    "p_ct_cP1 (Plasma IgA)": "#def9ea",
    "p_ct_cP2 (Plasma IgG)": "#fbeee1",
    "p_ct_cP3 (Plasma IgG prolif)": "#ebdad8",
    "p_ct_cS01 (Endo arterial)": "#def9ed",
    "p_ct_cS02 (Endo capillary)": "#def9ee",
    "p_ct_cS03 (Endo capillary)": "#eff9fa",
    "p_ct_cS04 (Endo)": "#e3ecf1",
    "p_ct_cS05 (Endo venous)": "#def9f1",
    "p_ct_cS06 (Endo lymphatic)": "#def9f3",
    "p_ct_cS07 (Endo capillary-like)": "#def9f4",
    "p_ct_cS08 (Endo arterial-like)": "#def9f5",
    "p_ct_cS09 (Endo)": "#def9f6",
    "p_ct_cS10 (Endo tip cells)": "#def9f7",
    "p_ct_cS11 (Endo proif)": "#def9f8",
    "p_ct_cS12 (Endo)": "#def9f9",
    "p_ct_cS13 (Endo venous-like)": "#def8f9",
    "p_ct_cS14 (Endo)": "#e6eff4",
    "p_ct_cS15 (Pericyte)": "#def6f9",
    "p_ct_cS16 (Pericyte)": "#def5f9",
    "p_ct_cS17 (Pericyte)": "#def4f9",
    "p_ct_cS18 (Pericyte)": "#def3f9",
    "p_ct_cS19 (Pericyte)": "#def2f9",
    "p_ct_cS20 (Pericyte prolif)": "#def1f9",
    "p_ct_cS21 (Fibro stem cell niche)": "#def0f9",
    "p_ct_cS22 (Fibro stem cell niche)": "#deeff9",
    "p_ct_cS23 (Fibro BMP-producing)": "#deeef9",
    "p_ct_cS24 (Fibro BMP-producing)": "#deedf9",
    "p_ct_cS25 (Fibro CCL8+)": "#deebf9",
    "p_ct_cS26 (Myofibro)": "#deeaf9",
    "p_ct_cS27 (CXCL14+ CAF)": "#dee9f9",
    "p_ct_cS28 (GREM1+ CAF)": "#dee8f9",
    "p_ct_cS29 (MMP3+ CAF)": "#dee7f9",
    "p_ct_cS30 (CAF CCL8 Fibro-like)": "#dee6f9",
    "p_ct_cS31 (CAF stem niche Fibro-like)": "#dee5f9",
    "p_ct_cS32 (Smooth Muscle)": "#dee4f9",
    "p_ct_cS33 (Schwann)": "#dee3f9",
    "p_ct_cTNI01 (CD4+ IL7R+)": "#dee2f9",
    "p_ct_cTNI02 (CD4+ IL7R+SELL+)": "#dee1f9",
    "p_ct_cTNI03 (CD4+ IL7R+HSP+)": "#dee0f9",
    "p_ct_cTNI04 (CD4+ IL7R+CCL5+)": "#dedff9",
    "p_ct_cTNI05 (CD4+ IL17+)": "#dedef9",
    "p_ct_cTNI06 (CD4+ TFH)": "#dfdef9",
    "p_ct_cTNI07 (CD4+ CXCL13+)": "#e0def9",
    "p_ct_cTNI08 (CD4+ Treg)": "#e1def9",
    "p_ct_cTNI09 (CD4+ Treg prolif)": "#e2def9",
    "p_ct_cTNI10 (CD8+ IL7R+)": "#e3def9",
    "p_ct_cTNI11 (CD8+GZMK+)": "#e4def9",
    "p_ct_cTNI12 (CD8+ IL7R+)": "#e5def9",
    "p_ct_cTNI13 (CD8+ T IL17+)": "#e6def9",
    "p_ct_cTNI14 (CD8+ CXCL13+)": "#e7def9",
    "p_ct_cTNI15 (CD8+ CXCL13+ HSP+)": "#e8def9",
    "p_ct_cTNI16 (CD8+ CXCL13+ prolif)": "#e9def9",
    "p_ct_cTNI17 (gd-like T)": "#eadef9",
    "p_ct_cTNI18 (gd-like T PDCD1+)": "#ebdef9",
    "p_ct_cTNI19 (gd-like T prolif)": "#ecdef9",
    "p_ct_cTNI20 (PLZF+ T)": "#eddef9",
    "p_ct_cTNI21 (PLZF+ T prolif)": "#eedef9",
    "p_ct_cTNI22 (cTNI22)": "#efdef9",
    "p_ct_cTNI23 (NK CD16A+)": "#f0def9",
    "p_ct_cTNI24 (NK GZMK+)": "#f1def9",
    "p_ct_cTNI25 (NK XCL1+)": "#f2def9",
    "p_ct_cTNI26 (ILC3)": "#f3def9",
    "ambiguous": "#f4f4f4",
    "filtered": "#e0e1e0",
}

colormap_coloncatlas_98_dark = {
    "p_ct_Tumor cE01 (Stem/TA-like)": "#ff6666",
    "p_ct_Tumor cE02 (Stem/TA-like/Immature Goblet)": "#ff7866",
    "p_ct_Tumor cE03 (Stem/TA-like prolif)": "#e2eecc",
    "p_ct_Tumor cE04 (Enterocyte 1)": "#ff9c66",
    "p_ct_Tumor cE05 (Enterocyte 2)": "#c47b8a",
    "p_ct_Tumor cE06 (Immature Goblet)": "#ffc166",
    "p_ct_Tumor cE07 (Goblet/Enterocyte)": "#ffd366",
    "p_ct_Tumor cE08 (Goblet)": "#ffe566",
    "p_ct_Tumor cE09 (Best4)": "#fff766",
    "p_ct_Tumor cE10 (Tuft)": "#f1ff66",
    "p_ct_Tumor cE11 (Enteroendocrine)": "#fd423c",
    "p_ct_cB1 (B IGD+IgM+)": "#8bdc3a",
    "p_ct_cB2 (B GC-like)": "#85dc3a",
    "p_ct_cB3 (B CD40+ GC-like)": "#7fdc3a",
    "p_ct_cE01 (Stem/TA-like)": "#79dc3a",
    "p_ct_cE02 (Stem/TA-like/Immature Goblet)": "#73dc3a",
    "p_ct_cE03 (Stem/TA-like prolif)": "#6cdc3a",
    "p_ct_cE04 (Enterocyte 1)": "#66dc3a",
    "p_ct_cE05 (Enterocyte 2)": "#60dc3a",
    "p_ct_cE06 (Immature Goblet)": "#5adc3a",
    "p_ct_cE07 (Goblet/Enterocyte)": "#54dc3a",
    "p_ct_cE08 (Goblet)": "#4edc3a",
    "p_ct_cE09 (Best4)": "#48dc3a",
    "p_ct_cE10 (Tuft)": "#42dc3a",
    "p_ct_cE11 (Enteroendocrine)": "#3cdc3a",
    "p_ct_cM01 (Monocyte)": "#00704b",
    "p_ct_cM02 (Macrophage-like)": "#85ba48",
    "p_ct_cM03 (DC1)": "#3adc4e",
    "p_ct_cM04 (DC2)": "#3adc54",
    "p_ct_cM05 (DC2 C1Q+)": "#3adc5a",
    "p_ct_cM06 (DC IL22RA2)": "#3adc60",
    "p_ct_cM07 (pDC)": "#3adc66",
    "p_ct_cM08 (AS-DC)": "#3adc6c",
    "p_ct_cM09 (mregDC)": "#3adc72",
    "p_ct_cM10 (Granulocyte)": "#3adc78",
    "p_ct_cMA01 (Mast)": "#3adc7f",
    "p_ct_cP1 (Plasma IgA)": "#3adc85",
    "p_ct_cP2 (Plasma IgG)": "#e7994f",
    "p_ct_cP3 (Plasma IgG prolif)": "#872215",
    "p_ct_cS01 (Endo arterial)": "#3adc97",
    "p_ct_cS02 (Endo capillary)": "#3adc9d",
    "p_ct_cS03 (Endo capillary)": "#a2dce5",
    "p_ct_cS04 (Endo)": "#5b8eac",
    "p_ct_cS05 (Endo venous)": "#3adcaf",
    "p_ct_cS06 (Endo lymphatic)": "#3adcb7",
    "p_ct_cS07 (Endo capillary-like)": "#3adcbd",
    "p_ct_cS08 (Endo arterial-like)": "#3adcc3",
    "p_ct_cS09 (Endo)": "#3adcc9",
    "p_ct_cS10 (Endo tip cells)": "#3adccf",
    "p_ct_cS11 (Endo proif)": "#3adcd5",
    "p_ct_cS12 (Endo)": "#3adcdb",
    "p_ct_cS13 (Endo venous-like)": "#3ad6dc",
    "p_ct_cS14 (Endo)": "#6da2be",
    "p_ct_cS15 (Pericyte)": "#3acadc",
    "p_ct_cS16 (Pericyte)": "#3ac3dc",
    "p_ct_cS17 (Pericyte)": "#3abddc",
    "p_ct_cS18 (Pericyte)": "#3ab7dc",
    "p_ct_cS19 (Pericyte)": "#3ab1dc",
    "p_ct_cS20 (Pericyte prolif)": "#3aabdc",
    "p_ct_cS21 (Fibro stem cell niche)": "#3aa5dc",
    "p_ct_cS22 (Fibro stem cell niche)": "#3a9fdc",
    "p_ct_cS23 (Fibro BMP-producing)": "#3a99dc",
    "p_ct_cS24 (Fibro BMP-producing)": "#3a93dc",
    "p_ct_cS25 (Fibro CCL8+)": "#3a8bdc",
    "p_ct_cS26 (Myofibro)": "#3a85dc",
    "p_ct_cS27 (CXCL14+ CAF)": "#3a7fdc",
    "p_ct_cS28 (GREM1+ CAF)": "#3a79dc",
    "p_ct_cS29 (MMP3+ CAF)": "#3a73dc",
    "p_ct_cS30 (CAF CCL8 Fibro-like)": "#3a6ddc",
    "p_ct_cS31 (CAF stem niche Fibro-like)": "#3a67dc",
    "p_ct_cS32 (Smooth Muscle)": "#3a61dc",
    "p_ct_cS33 (Schwann)": "#3a5bdc",
    "p_ct_cTNI01 (CD4+ IL7R+)": "#3a55dc",
    "p_ct_cTNI02 (CD4+ IL7R+SELL+)": "#3a4edc",
    "p_ct_cTNI03 (CD4+ IL7R+HSP+)": "#3a48dc",
    "p_ct_cTNI04 (CD4+ IL7R+CCL5+)": "#3a42dc",
    "p_ct_cTNI05 (CD4+ IL17+)": "#3a3cdc",
    "p_ct_cTNI06 (CD4+ TFH)": "#403adc",
    "p_ct_cTNI07 (CD4+ CXCL13+)": "#463adc",
    "p_ct_cTNI08 (CD4+ Treg)": "#4c3adc",
    "p_ct_cTNI09 (CD4+ Treg prolif)": "#523adc",
    "p_ct_cTNI10 (CD8+ IL7R+)": "#583adc",
    "p_ct_cTNI11 (CD8+GZMK+)": "#5e3adc",
    "p_ct_cTNI12 (CD8+ IL7R+)": "#663adc",
    "p_ct_cTNI13 (CD8+ T IL17+)": "#6c3adc",
    "p_ct_cTNI14 (CD8+ CXCL13+)": "#723adc",
    "p_ct_cTNI15 (CD8+ CXCL13+ HSP+)": "#783adc",
    "p_ct_cTNI16 (CD8+ CXCL13+ prolif)": "#7e3adc",
    "p_ct_cTNI17 (gd-like T)": "#843adc",
    "p_ct_cTNI18 (gd-like T PDCD1+)": "#8a3adc",
    "p_ct_cTNI19 (gd-like T prolif)": "#903adc",
    "p_ct_cTNI20 (PLZF+ T)": "#963adc",
    "p_ct_cTNI21 (PLZF+ T prolif)": "#9d3adc",
    "p_ct_cTNI22 (cTNI22)": "#a33adc",
    "p_ct_cTNI23 (NK CD16A+)": "#a93adc",
    "p_ct_cTNI24 (NK GZMK+)": "#af3adc",
    "p_ct_cTNI25 (NK XCL1+)": "#b53adc",
    "p_ct_cTNI26 (ILC3)": "#bb3adc",
    "ambiguous": "#bdbdbd",
    "filtered": "#464c46",
}


#############column order ##################

crc98_order_col = [
    "p_ct_Tumor cE01 (Stem/TA-like)",
    "p_ct_Tumor cE02 (Stem/TA-like/Immature Goblet)",
    "p_ct_Tumor cE03 (Stem/TA-like prolif)",
    "p_ct_Tumor cE04 (Enterocyte 1)",
    "p_ct_Tumor cE05 (Enterocyte 2)",
    "p_ct_Tumor cE06 (Immature Goblet)",
    "p_ct_Tumor cE07 (Goblet/Enterocyte)",
    "p_ct_Tumor cE08 (Goblet)",
    "p_ct_Tumor cE09 (Best4)",
    "p_ct_Tumor cE10 (Tuft)",
    "p_ct_Tumor cE11 (Enteroendocrine)",
    "p_ct_cE01 (Stem/TA-like)",
    "p_ct_cE02 (Stem/TA-like/Immature Goblet)",
    "p_ct_cE03 (Stem/TA-like prolif)",
    "p_ct_cE04 (Enterocyte 1)",
    "p_ct_cE05 (Enterocyte 2)",
    "p_ct_cE06 (Immature Goblet)",
    "p_ct_cE07 (Goblet/Enterocyte)",
    "p_ct_cE08 (Goblet)",
    "p_ct_cE09 (Best4)",
    "p_ct_cE10 (Tuft)",
    "p_ct_cE11 (Enteroendocrine)",
    "p_ct_cM01 (Monocyte)",
    "p_ct_cM02 (Macrophage-like)",
    "p_ct_cM03 (DC1)",
    "p_ct_cM04 (DC2)",
    "p_ct_cM05 (DC2 C1Q+)",
    "p_ct_cM06 (DC IL22RA2)",
    "p_ct_cM07 (pDC)",
    "p_ct_cM08 (AS-DC)",
    "p_ct_cM09 (mregDC)",
    "p_ct_cM10 (Granulocyte)",
    "p_ct_cMA01 (Mast)",
    "p_ct_cB1 (B IGD+IgM+)",
    "p_ct_cB2 (B GC-like)",
    "p_ct_cB3 (B CD40+ GC-like)",
    "p_ct_cP1 (Plasma IgA)",
    "p_ct_cP2 (Plasma IgG)",
    "p_ct_cP3 (Plasma IgG prolif)",
    "p_ct_cTNI01 (CD4+ IL7R+)",
    "p_ct_cTNI02 (CD4+ IL7R+SELL+)",
    "p_ct_cTNI03 (CD4+ IL7R+HSP+)",
    "p_ct_cTNI04 (CD4+ IL7R+CCL5+)",
    "p_ct_cTNI05 (CD4+ IL17+)",
    "p_ct_cTNI06 (CD4+ TFH)",
    "p_ct_cTNI07 (CD4+ CXCL13+)",
    "p_ct_cTNI08 (CD4+ Treg)",
    "p_ct_cTNI09 (CD4+ Treg prolif)",
    "p_ct_cTNI10 (CD8+ IL7R+)",
    "p_ct_cTNI11 (CD8+GZMK+)",
    "p_ct_cTNI12 (CD8+ IL7R+)",
    "p_ct_cTNI13 (CD8+ T IL17+)",
    "p_ct_cTNI14 (CD8+ CXCL13+)",
    "p_ct_cTNI15 (CD8+ CXCL13+ HSP+)",
    "p_ct_cTNI16 (CD8+ CXCL13+ prolif)",
    "p_ct_cTNI17 (gd-like T)",
    "p_ct_cTNI18 (gd-like T PDCD1+)",
    "p_ct_cTNI19 (gd-like T prolif)",
    "p_ct_cTNI20 (PLZF+ T)",
    "p_ct_cTNI21 (PLZF+ T prolif)",
    "p_ct_cTNI22 (cTNI22)",
    "p_ct_cTNI23 (NK CD16A+)",
    "p_ct_cTNI24 (NK GZMK+)",
    "p_ct_cTNI25 (NK XCL1+)",
    "p_ct_cTNI26 (ILC3)",
    "p_ct_cS01 (Endo arterial)",
    "p_ct_cS02 (Endo capillary)",
    "p_ct_cS03 (Endo capillary)",
    "p_ct_cS04 (Endo)",
    "p_ct_cS05 (Endo venous)",
    "p_ct_cS06 (Endo lymphatic)",
    "p_ct_cS07 (Endo capillary-like)",
    "p_ct_cS08 (Endo arterial-like)",
    "p_ct_cS09 (Endo)",
    "p_ct_cS10 (Endo tip cells)",
    "p_ct_cS11 (Endo proif)",
    "p_ct_cS12 (Endo)",
    "p_ct_cS13 (Endo venous-like)",
    "p_ct_cS14 (Endo)",
    "p_ct_cS15 (Pericyte)",
    "p_ct_cS16 (Pericyte)",
    "p_ct_cS17 (Pericyte)",
    "p_ct_cS18 (Pericyte)",
    "p_ct_cS19 (Pericyte)",
    "p_ct_cS20 (Pericyte prolif)",
    "p_ct_cS21 (Fibro stem cell niche)",
    "p_ct_cS22 (Fibro stem cell niche)",
    "p_ct_cS23 (Fibro BMP-producing)",
    "p_ct_cS24 (Fibro BMP-producing)",
    "p_ct_cS25 (Fibro CCL8+)",
    "p_ct_cS26 (Myofibro)",
    "p_ct_cS27 (CXCL14+ CAF)",
    "p_ct_cS28 (GREM1+ CAF)",
    "p_ct_cS29 (MMP3+ CAF)",
    "p_ct_cS30 (CAF CCL8 Fibro-like)",
    "p_ct_cS31 (CAF stem niche Fibro-like)",
    "p_ct_cS32 (Smooth Muscle)",
    "p_ct_cS33 (Schwann)",
    "ambiguous",
    "filtered",
]

crc98_major_col = {
    "TumorEpi": [
        "p_ct_Tumor cE01 (Stem/TA-like)",
        "p_ct_Tumor cE02 (Stem/TA-like/Immature Goblet)",
        "p_ct_Tumor cE03 (Stem/TA-like prolif)",
        "p_ct_Tumor cE04 (Enterocyte 1)",
        "p_ct_Tumor cE05 (Enterocyte 2)",
        "p_ct_Tumor cE06 (Immature Goblet)",
        "p_ct_Tumor cE07 (Goblet/Enterocyte)",
        "p_ct_Tumor cE08 (Goblet)",
        "p_ct_Tumor cE09 (Best4)",
        "p_ct_Tumor cE10 (Tuft)",
        "p_ct_Tumor cE11 (Enteroendocrine)",
    ],
    "NormalEpi": [
        "p_ct_cE01 (Stem/TA-like)",
        "p_ct_cE02 (Stem/TA-like/Immature Goblet)",
        "p_ct_cE03 (Stem/TA-like prolif)",
        "p_ct_cE04 (Enterocyte 1)",
        "p_ct_cE05 (Enterocyte 2)",
        "p_ct_cE06 (Immature Goblet)",
        "p_ct_cE07 (Goblet/Enterocyte)",
        "p_ct_cE08 (Goblet)",
        "p_ct_cE09 (Best4)",
        "p_ct_cE10 (Tuft)",
        "p_ct_cE11 (Enteroendocrine)",
    ],
    "Myeloid": [
        "p_ct_cMA01 (Mast)",
        "p_ct_cM01 (Monocyte)",
        "p_ct_cM02 (Macrophage-like)",
        "p_ct_cM03 (DC1)",
        "p_ct_cM04 (DC2)",
        "p_ct_cM05 (DC2 C1Q+)",
        "p_ct_cM06 (DC IL22RA2)",
        "p_ct_cM07 (pDC)",
        "p_ct_cM08 (AS-DC)",
        "p_ct_cM09 (mregDC)",
        "p_ct_cM10 (Granulocyte)",
    ],
    "B": [
        "p_ct_cB1 (B IGD+IgM+)",
        "p_ct_cB2 (B GC-like)",
        "p_ct_cB3 (B CD40+ GC-like)",
        "p_ct_cP1 (Plasma IgA)",
        "p_ct_cP2 (Plasma IgG)",
        "p_ct_cP3 (Plasma IgG prolif)",
    ],
    "T": [
        "p_ct_cTNI01 (CD4+ IL7R+)",
        "p_ct_cTNI02 (CD4+ IL7R+SELL+)",
        "p_ct_cTNI03 (CD4+ IL7R+HSP+)",
        "p_ct_cTNI04 (CD4+ IL7R+CCL5+)",
        "p_ct_cTNI05 (CD4+ IL17+)",
        "p_ct_cTNI06 (CD4+ TFH)",
        "p_ct_cTNI07 (CD4+ CXCL13+)",
        "p_ct_cTNI08 (CD4+ Treg)",
        "p_ct_cTNI09 (CD4+ Treg prolif)",
        "p_ct_cTNI10 (CD8+ IL7R+)",
        "p_ct_cTNI11 (CD8+GZMK+)",
        "p_ct_cTNI12 (CD8+ IL7R+)",
        "p_ct_cTNI13 (CD8+ T IL17+)",
        "p_ct_cTNI14 (CD8+ CXCL13+)",
        "p_ct_cTNI15 (CD8+ CXCL13+ HSP+)",
        "p_ct_cTNI16 (CD8+ CXCL13+ prolif)",
        "p_ct_cTNI17 (gd-like T)",
        "p_ct_cTNI18 (gd-like T PDCD1+)",
        "p_ct_cTNI19 (gd-like T prolif)",
        "p_ct_cTNI20 (PLZF+ T)",
        "p_ct_cTNI21 (PLZF+ T prolif)",
        "p_ct_cTNI22 (cTNI22)",
        "p_ct_cTNI23 (NK CD16A+)",
        "p_ct_cTNI24 (NK GZMK+)",
        "p_ct_cTNI25 (NK XCL1+)",
        "p_ct_cTNI26 (ILC3)",
    ],
    "EndoStromal": [
        "p_ct_cS01 (Endo arterial)",
        "p_ct_cS02 (Endo capillary)",
        "p_ct_cS03 (Endo capillary)",
        "p_ct_cS04 (Endo)",
        "p_ct_cS05 (Endo venous)",
        "p_ct_cS06 (Endo lymphatic)",
        "p_ct_cS07 (Endo capillary-like)",
        "p_ct_cS08 (Endo arterial-like)",
        "p_ct_cS09 (Endo)",
        "p_ct_cS10 (Endo tip cells)",
        "p_ct_cS11 (Endo proif)",
        "p_ct_cS12 (Endo)",
        "p_ct_cS13 (Endo venous-like)",
        "p_ct_cS14 (Endo)",
        "p_ct_cS15 (Pericyte)",
        "p_ct_cS16 (Pericyte)",
        "p_ct_cS17 (Pericyte)",
        "p_ct_cS18 (Pericyte)",
        "p_ct_cS19 (Pericyte)",
        "p_ct_cS20 (Pericyte prolif)",
        "p_ct_cS21 (Fibro stem cell niche)",
        "p_ct_cS22 (Fibro stem cell niche)",
        "p_ct_cS23 (Fibro BMP-producing)",
        "p_ct_cS24 (Fibro BMP-producing)",
        "p_ct_cS25 (Fibro CCL8+)",
        "p_ct_cS26 (Myofibro)",
        "p_ct_cS27 (CXCL14+ CAF)",
        "p_ct_cS28 (GREM1+ CAF)",
        "p_ct_cS29 (MMP3+ CAF)",
        "p_ct_cS30 (CAF CCL8 Fibro-like)",
        "p_ct_cS31 (CAF stem niche Fibro-like)",
        "p_ct_cS32 (Smooth Muscle)",
        "p_ct_cS33 (Schwann)",
    ],
    "filtered": [
        "ambiguous",
        "filtered",
    ],
    ### separate types for dotplot
    "Endo": [
        "p_ct_cS01 (Endo arterial)",
        "p_ct_cS02 (Endo capillary)",
        "p_ct_cS03 (Endo capillary)",
        "p_ct_cS04 (Endo)",
        "p_ct_cS05 (Endo venous)",
        "p_ct_cS06 (Endo lymphatic)",
        "p_ct_cS07 (Endo capillary-like)",
        "p_ct_cS08 (Endo arterial-like)",
        "p_ct_cS09 (Endo)",
        "p_ct_cS10 (Endo tip cells)",
        "p_ct_cS11 (Endo proif)",
        "p_ct_cS12 (Endo)",
        "p_ct_cS13 (Endo venous-like)",
        "p_ct_cS14 (Endo)",
    ],
    "Peri": [
        "p_ct_cS15 (Pericyte)",
        "p_ct_cS16 (Pericyte)",
        "p_ct_cS17 (Pericyte)",
        "p_ct_cS18 (Pericyte)",
        "p_ct_cS19 (Pericyte)",
        "p_ct_cS20 (Pericyte prolif)",
    ],
    "Fibro": [
        "p_ct_cS21 (Fibro stem cell niche)",
        "p_ct_cS22 (Fibro stem cell niche)",
        "p_ct_cS23 (Fibro BMP-producing)",
        "p_ct_cS24 (Fibro BMP-producing)",
        "p_ct_cS25 (Fibro CCL8+)",
        "p_ct_cS26 (Myofibro)",
        "p_ct_cS27 (CXCL14+ CAF)",
        "p_ct_cS28 (GREM1+ CAF)",
        "p_ct_cS29 (MMP3+ CAF)",
        "p_ct_cS30 (CAF CCL8 Fibro-like)",
        "p_ct_cS31 (CAF stem niche Fibro-like)",
        #'p_ct_cS32 (Smooth Muscle)',
        #'p_ct_cS33 (Schwann)',
    ],
    "CD4T": [
        "p_ct_cTNI01 (CD4+ IL7R+)",
        "p_ct_cTNI02 (CD4+ IL7R+SELL+)",
        "p_ct_cTNI03 (CD4+ IL7R+HSP+)",
        "p_ct_cTNI04 (CD4+ IL7R+CCL5+)",
        "p_ct_cTNI05 (CD4+ IL17+)",
        "p_ct_cTNI06 (CD4+ TFH)",
        "p_ct_cTNI07 (CD4+ CXCL13+)",
        "p_ct_cTNI08 (CD4+ Treg)",
        "p_ct_cTNI09 (CD4+ Treg prolif)",
    ],
    "CD8T": [
        "p_ct_cTNI10 (CD8+ IL7R+)",
        "p_ct_cTNI11 (CD8+GZMK+)",
        "p_ct_cTNI12 (CD8+ IL7R+)",
        "p_ct_cTNI13 (CD8+ T IL17+)",
        "p_ct_cTNI14 (CD8+ CXCL13+)",
        "p_ct_cTNI15 (CD8+ CXCL13+ HSP+)",
        "p_ct_cTNI16 (CD8+ CXCL13+ prolif)",
    ],
    "Tother": [
        "p_ct_cTNI17 (gd-like T)",
        "p_ct_cTNI18 (gd-like T PDCD1+)",
        "p_ct_cTNI19 (gd-like T prolif)",
        "p_ct_cTNI20 (PLZF+ T)",
        "p_ct_cTNI21 (PLZF+ T prolif)",
        "p_ct_cTNI22 (cTNI22)",
        "p_ct_cTNI23 (NK CD16A+)",
        "p_ct_cTNI24 (NK GZMK+)",
        "p_ct_cTNI25 (NK XCL1+)",
        "p_ct_cTNI26 (ILC3)",
    ],
}


crc98_epicol = [
    "p_ct_cE01 (Stem/TA-like)",
    "p_ct_Tumor cE01 (Stem/TA-like)",
    "p_ct_cE02 (Stem/TA-like/Immature Goblet)",
    "p_ct_Tumor cE02 (Stem/TA-like/Immature Goblet)",
    "p_ct_cE03 (Stem/TA-like prolif)",
    "p_ct_Tumor cE03 (Stem/TA-like prolif)",
    "p_ct_cE04 (Enterocyte 1)",
    "p_ct_Tumor cE04 (Enterocyte 1)",
    "p_ct_cE05 (Enterocyte 2)",
    "p_ct_Tumor cE05 (Enterocyte 2)",
    "p_ct_cE06 (Immature Goblet)",
    "p_ct_Tumor cE06 (Immature Goblet)",
    "p_ct_cE07 (Goblet/Enterocyte)",
    "p_ct_Tumor cE07 (Goblet/Enterocyte)",
    "p_ct_cE08 (Goblet)",
    "p_ct_Tumor cE08 (Goblet)",
    "p_ct_cE09 (Best4)",
    "p_ct_Tumor cE09 (Best4)",
    "p_ct_cE10 (Tuft)",
    "p_ct_Tumor cE10 (Tuft)",
    "p_ct_cE11 (Enteroendocrine)",
    "p_ct_Tumor cE11 (Enteroendocrine)",
]

crc98_normalepicol = [
    "p_ct_cE01 (Stem/TA-like)",
    "p_ct_cE02 (Stem/TA-like/Immature Goblet)",
    "p_ct_cE03 (Stem/TA-like prolif)",
    "p_ct_cE04 (Enterocyte 1)",
    "p_ct_cE05 (Enterocyte 2)",
    "p_ct_cE06 (Immature Goblet)",
    "p_ct_cE07 (Goblet/Enterocyte)",
    "p_ct_cE08 (Goblet)",
    "p_ct_cE09 (Best4)",
    "p_ct_cE10 (Tuft)",
    "p_ct_cE11 (Enteroendocrine)",
]
crc98_tumorepicol = [
    "p_ct_Tumor cE01 (Stem/TA-like)",
    "p_ct_Tumor cE02 (Stem/TA-like/Immature Goblet)",
    "p_ct_Tumor cE03 (Stem/TA-like prolif)",
    "p_ct_Tumor cE04 (Enterocyte 1)",
    "p_ct_Tumor cE05 (Enterocyte 2)",
    "p_ct_Tumor cE06 (Immature Goblet)",
    "p_ct_Tumor cE07 (Goblet/Enterocyte)",
    "p_ct_Tumor cE08 (Goblet)",
    "p_ct_Tumor cE09 (Best4)",
    "p_ct_Tumor cE10 (Tuft)",
    "p_ct_Tumor cE11 (Enteroendocrine)",
]


################ gene marker #################

#'PHGR1','GAPDH','PTMA','ACTB',
marker_crc98_dict = {
    "NormalEpi": [
        "LEFTY1",
        "EPHB3",
        "LGR5",
        "ALDH1B1",
        "CDCA7",
        "OLFM4",
        "STMN1",
        "HMGB2",
        "MKI67",
        "CLCA4",
        "AQP8",
        "TMIGD1",
        "GUCA2B",
        "SULT1A2",
        "SLC26A3",
        "SLC26A2",
        "CA1",
        "SPINK1",
        "CLCA1",
        "ITLN1",
        "LRRC26",
        "RETNLB",
        "SPINK4",
        "FCGBP",
        "MUC2",
        "ZG16",
        "BCAS1",
        "ENTPD8",
        "MUC1",
        "TFF1",
        "S100P",
        "NOTCH2NL",
        "OTOP3",
        "REN",
        "BEST4",
        "OTOP2",
        "CA7",
        "AZGP1",
        "HCK",
        "LRMP",
        "SH2D6",
        "HPGDS",
        "PTGS1",
        "CHGA",
        "CRYBA2",
        "SCGN",
        "PYY",
        "SCT",
        "INSL5",
    ],
    "CD4T": [
        "IL7R",
        "LEF1",
        "SELL",
        "CCR7",
        "TCF7",
        "HSPA1A",
        "HSPA1B",
        "CCL5",
        "CD40LG",
        "KLRB1",
        "RORA",
        "IL17A",
        "ICA1",
        "IL6ST",
        "PASK",
        "BCL6",
        "CXCR5",
        "ZBED2",
        "CXCL13",
        "TNFRSF18",
        "CTLA4",
        "TNFRSF4",
        "BATF",
        "FOXP3",
        "IL2RA",
        "IL1RL1",
        "SOX4",
        "STMN1",
        "HMGB2",
        "MKI67",
    ],
    "CD8T": [
        "IL7R",
        "TCF7",
        "GZMK",
        "RORA",
        "KLRB1",
        "IL17A",
        "TOX",
        "ITGB7",
        "ITGAE",
        "RBPJ",
        "ETV1",
        "PDCD1",
        "LAG3",
        "HAVCR2",
        "TIGIT",
        "IFNG",
        "PRF1",
        "GNLY",
        "GZMH",
        "GZMB",
        "CXCL13",
        "TNFSF4",
        "ENTPD1",
        "ZBED2",
        "HSPA1A",
        "HSPA1B",
        "STMN1",
        "HMGB2",
        "MKI67",
    ],
    "Tother": [
        "TRGC1",
        "TRGC2",
        "KLRC2",
        "LAG3",
        "HAVCR2",
        "PDCD1",
        "GZMB",
        "GNLY",
        "PRF1",
        "ZBTB16",
        "TYROBP",
        "FCER1G",
        "STMN1",
        "HMGB2",
        "MK167",
        "CD8A",
        "CD8B",
        "KLF2",
        "GZMH",
        "LYAR",
        "KLRG1",
        "B3GAT1",
        "TBX21",
        "FGFBP2",
        "CX3CR1",
        "FCGR3A",
        "KLRF1",
        "CMC1",
        "KLRC1",
        "NCAM1",
        "XCL1",
        "XCL2",
        "SOX4",
        "KLRB1",
        "BCL6",
        "LST1",
        "RORC",
    ],
    "B": [
        "CD79A",
        "CD74",
        "MS4A1",
        "CXCR5",
        "CD19",
        "TCL1A",
        "FCER2",
        "IGHD",
        "IGHM",
        "TNFRSF13B",
        "GPR183",
        "CD40",
        "CD86",
        "CD80",
        "MARCKSL1",
        "RGS13",
        "AICDA",
        "CD27",
        "IGHA1",
        "SDC1",
        "TNFRSF17",
        "JCHAIN",
        "MZB1",
        "DERL3",
        "CD38",
        "IGHG1",
        "IGHG3",
        "IGHG4",
        "STMN1",
        "HMGB2",
        "MKI67",
    ],
    "Myeloid": [
        "TPSAB1",
        "CPA3",
        "KIT",
        "FCN1",
        "VCAN",
        "CD300E",
        "S100A12",
        "CD163",
        "APOC1",
        "LIPA",
        "C1QA",
        "CSF1R",
        "CLEC9A",
        "XCR1",
        "BATF3",
        "FCER1A",
        "CD1C",
        "CLEC10A",
        "CD1E",
        "PAK1",
        "IL22RA2",
        "CLEC4C",
        "IRF7",
        "CXCR3",
        "LILRA4",
        "SIGLEC6",
        "PPP1R14A",
        "AXL",
        "LAMP3",
        "CCL19",
        "CCL22",
        "CCR7",
        "CD274",
        "FCGR3B",
        "HCAR2",
        "GPR109B",
        "CSF3R",
    ],
    "Endo": [
        "LTC4S",
        "PCSK5",
        "BMX",
        "SEMA3G",
        "HEY1",
        "ACE",
        "CD320",
        "ENPP2",
        "SLC14A1",
        "TMEM88",
        "CD36",
        "CA4",
        "PASK",
        "HIGD1B",
        "COX412",
        "NOTCH3",
        "PDGFRB",
        "RGS5",
        "MYL9",
        "ACTA2",
        "MCAM",
        "NID1",
        "NID2",
        "CPE",
        "ADGRG6",
        "MADCAM1",
        "ACKR1",
        "IL1R1",
        "CCL14",
        "CCL23",
        "CCL21",
        "LYVE1",
        "MMRN1",
        "RBP7",
        "PPARG",
        "FABP4",
        "BTNL9",
        "CD300LG",
        "GJA4",
        "GJA5",
        "VEGFC",
        "HECW2",
        "GRB10",
        "CA2",
        "ANGPT2",
        "ESMT",
        "APLN",
        "PDGFB",
        "PGF",
        "CXCR4",
        "STMN1",
        "HMGN2",
        "HMGB2",
        "MK167",
        "CSF3",
        "HIF1A",
        "S100A3",
        "SELE",
        "SELP",
        "IL33",
        "VCAM1",
    ],
    "Peri": [
        "RGS5",
        "ADAMDEC1",
        "COX412",
        "HIGD1B",
        "CD36",
        "STEAP4",
        "APOE",
        "FABP4",
        "MT1M",
        "SYNPO2",
        "HOPX",
        "SMOC2",
        "MYH11",
        "MIR",
        "143HG",
        "MAP3K7CL",
        "NOTCH3",
        "MYLK",
        "ACTA2",
        "GJA4",
        "MMP11",
        "LURAP1L",
        "NDUFA4L2",
        "KCNJ8",
        "PLXDC1",
        "EDNRA",
        "CCDC102B",
        "HIF1A",
        "ERRFI1",
        "COL15A1",
        "PRRX1",
        "SULF1",
        "TSPAN15",
        "EMP1",
        "IL32",
        "GOS2",
        "VCAN",
        "S100A10",
        "POSTN",
    ],
    "Fibro": [
        "LOXL1",
        "PCOLCE2",
        "SEMA3B",
        "CLEC3B",
        "TPPP3",
        "WNT2B",
        "MFAP5",
        "CD34",
        "SFRP2",
        "CXCL12",
        "RSPO3",
        "GREM1",
        "CXCL14",
        "COL4A5",
        "CXCR4",
        "WNT5B",
        "COL4A6",
        "BMP4",
        "SOX6",
        "FRZB",
        "ENHO",
        "BMP5",
        "BMP2",
        "CCL2",
        "CCL8",
        "CCL11",
        "CCL13",
        "TAGLN",
        "ACTA2",
        "MYH11",
        "PDPN",
        "INHBA",
        "DKK3",
        "THY1",
        "MMP11",
        "WNT4",
        "AGT",
        "COL11A1",
        "ADAM12",
        "CTHRC1",
        "FAP",
        "TNC",
        "COL7A1",
        "MMP1",
        "CXCL1",
        "CXCL5",
        "MMP3",
        "CXCL3",
        "CXCL8",
        "L24",
        "STC1",
        "ADAM28",
        "FABP5",
        "TFPI2",
        "PAPPA",
        "PTGDS",
        "CHRDL1",
        "NT5E",
        "IGF1",
        "PLA2G2A",
        "BOC",
        "LXN",
        "RARRES1",
    ],
}
colon_epithelial_marker = [
    "LEFTY1",
    "EPHB3",
    "LGR5",
    "ALDH1B1",
    "CDCA7",
    "OLFM4",
    "STMN1",
    "HMGB2",
    "MKI67",
    "CLCA4",
    "AQP8",
    "TMIGD1",
    "GUCA2B",
    "SULT1A2",
    "SLC26A3",
    "SLC26A2",
    "CA1",
    "SPINK1",
    "CLCA1",
    "ITLN1",
    "LRRC26",
    "RETNLB",
    "SPINK4",
    "FCGBP",
    "MUC2",
    "ZG16",
    "ENTPD8",
    "MUC1",
    "TFF1",
    "S100P",
    "OTOP3",
    "REN",
    "BEST4",
    "OTOP2",
    "CA7",
    "AZGP1",
    "HCK",
    "LRMP",
    "SH2D6",
    "HPGDS",
    "PTGS1",
    "CHGA",
    "CRYBA2",
    "SCGN",
    "PYY",
    "SCT",
    "INSL5",
]  # no NOTCH2NL, BCAS1 in ST data

colon_epithelial_marker_dict = {
    "Stem/TA-like": [
        "LEFTY1",
        "EPHB3",
        "LGR5",
        "ALDH1B1",
        "CDCA7",
        "OLFM4",
        "STMN1",
        "HMGB2",
        "MKI67",
    ],
    "Enterocyte": [
        "CLCA4",
        "AQP8",
        "TMIGD1",
        "GUCA2B",
        "SULT1A2",
        "SLC26A3",
        "SLC26A2",
        "CA1",
    ],
    "Goblet": [
        "SPINK1",
        "CLCA1",
        "ITLN1",
        "LRRC26",
        "RETNLB",
        "SPINK4",
        "FCGBP",
        "MUC2",
        "ZG16",
        "BCAS1",
        "ENTPD8",
        "MUC1",
        "TFF1",
        "S100P",
    ],
    "Best4": [
        "NOTCH2NL",
        "OTOP3",
        "REN",
        "BEST4",
        "OTOP2",
        "CA7",
    ],
    "Tuft": ["AZGP1", "HCK", "LRMP", "SH2D6", "HPGDS", "PTGS1"],
    "Enteroendocrine": ["CHGA", "CRYBA2", "SCGN", "PYY", "SCT", "INSL5"],
}

# marker: see https://www.nature.com/articles/s41588-022-01088-x
# regev group https://pdf.sciencedirectassets.com/272196/1-s2.0-S0092867418X00166/1-s2.0-S0092867419307329/main.pdf?X-Amz-Security-Token=IQoJb3JpZ2luX2VjEKj%2F%2F%2F%2F%2F%2F%2F%2F%2F%2FwEaCXVzLWVhc3QtMSJHMEUCIQC0M9TltK3%2FGqfzKVCoFTekhWxFDTq1kmt6GO%2Bw%2BzHb7AIgPnjjDMDI0wyi5PwJRj27%2B7lNuzylkfJL2%2BSQE4xmB8oquwUI4f%2F%2F%2F%2F%2F%2F%2F%2F%2F%2FARAFGgwwNTkwMDM1NDY4NjUiDNI6up%2BqTA2HVTJQ3iqPBWk2VVzzIz%2BNMfb8ulYVaxQV3Op5DpiiRondhVEWDLR5hx3G0K1cXpuBfDEax5kg%2FHx3M2dLrDLGwe34iHH0jwHeAPudcTMln388K%2FTAyiFwobN4IUqYpdgcMTnNjEzIZfkiXGrnFAOKnPb%2BCxNpHmOCIilhwxjwOhuYK4QKtw6q5JJITdm0Z5pqGHJf6nHTamzlR1nnv4w99iGfL1of6yPLlWkxIvA3BnlQ6pLAvApSDEA%2Fv1gjjJyneAI1PnoVkWU0facrKJ9V2H43xL%2BVqXudu5wBfSgvWGCpmdyhqKUm6Jx32GEP758d24aqSgXamGRjxioJST363X5lGf3r3hV1erh6TO2494LXFoKWkpojCFVGcS1vhaj7c0Vl2S4pfKRb6MH6ET2AnylrVhcW8VoqY1hiA10m3y3acwW4oAol49QaiukQsr9ZdGEi0qlWUSYeRxx2yjgyuv5WcDi0HCTzp1piqvu6cQrwK18tCXUuY3ttv5C2jLZVRL7EkpUWeL51w9U%2FGuPQVG%2FpnhwV%2B26djMM%2BCj9lvobEPy1Gv4x4hRO44%2BDTQO0fJA7MLlPB7wIboOSCc1oCCDvFoZzyxigl%2BVg958j5LHK9YNVspskN79HhEJxD%2FGkv7blDRRUkVEy59uId8%2BX3FGBwtV0f73fmHLsNcQi%2B5QJN1TPGumWunl8y0PmHx7k8l9w00CH2cQ69lIVIBvbOD6CMfP5olwHnmGbjB2reQ1w6iWn2JmD0z%2BTO6Ww8VxFY69RDzphtvp7qX5WvamdJaALqkZ6RAg%2Bx9uVX0ZxZI6YKUDUQQwKNQwa%2F4Mn2Tuo%2F5z4dGOO%2BJzIG4%2BALLRFn%2FacAVFJLGeByh0T5pDL1i4snOyELAV4wwb%2BBsQY6sQH2JeI%2Fof%2FpkjEgMSjfgAcev0tdZRrbZLM2IQJws7hK3ogs2XtcLse0qvS0b5ibT2yMWWEbr0F%2Bbpkh0tYY34GLbBmUUY3LCmwDc4PaK2BULUCDYi8LGu224Kt9NWz9oUGD2oadKwyHm2qRjbyQiU7ZMGRpv7TQhf05YR8sJTAQpimkcBiloP2WA5JKM%2FlMOrm2IyCOH8MF7aUbt4puy7D46dgVk5l9xd%2F5zNlzg%2F3s5RE%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Date=20240418T005238Z&X-Amz-SignedHeaders=host&X-Amz-Expires=300&X-Amz-Credential=ASIAQ3PHCVTY3K6HJILE%2F20240418%2Fus-east-1%2Fs3%2Faws4_request&X-Amz-Signature=f574f3d248efc5c3aa7bf43648b7522f90b46d4d65d4e6010214ab116b4f4c11&hash=41f0658a3a0552a478aaf0a832744ce3d8d4df99c8241178aad8f568d82641cd&host=68042c943591013ac2b2430a89b270f6af2c76d8dfd086a07176afe7c76c2c61&pii=S0092867419307329&tid=spdf-24a30430-2430-4ed3-b0cf-178094bf0b23&sid=3001fddd53ddd140fb7b77c36339939fbaabgxrqa&type=client&tsoh=d3d3LnNjaWVuY2VkaXJlY3QuY29t&ua=1315585657510e0d5106&rr=8760ac5b3e7a9e05&cc=us
markers_regev_epi = [
    "ASCL2",
    "LGR5",
    "LEFTY1",
    "CDC25C",
    "SLC26A2",
    "FABP1",
    "CA1",
    "EPCAM",
    "RBP2",
    "SLC26A3",
    "CCL20",
    "TNFRSF11A",
    "BEST4",
    "TFF1",
    "MUC2",
    "HCK",
    "TRPM5",
    "SCGN",
    "CHGA",
    "CCL7",
    "CCL13",
]
markers_regev_stroma = [
    "RSPO3",
    "CCL11",
    "WNT5B",
    "BMP4",
    "CHI3L1",
    "TAGLN",
    "ACTA2",
    "WNT2B",
    "SOX17",
    "RBP7",
    "VWA1",
    "PLVAP",
    "CDH5",
    "CD36",
    "MADCAM1",
    "DARC",
    "RGS5",
    "SOX10",
    "S100B",
    "CD68",
    "XCR1",
    "CLEC9A",
]
markers_regev_myeloid = [
    "CD1C",
    "S100A8",
    "S100A9",
    "TPSAB1",
    "OSM",
    "NKG7",
    "KLRC1",
    "CCR7",
]
markers_regev_t = ["CCR7", "FOXP3", "CTLA4", "CD8B", "CXCR6", "CD3D"]
markers_regev_b = ["MZB1", "IGHA1", "SELL", "CD19", "AICDA"]
colon_markers_regev_dict = {
    "Epi": [
        "ASCL2",
        "LGR5",
        "LEFTY1",
        "CDC25C",
        "SLC26A2",
        "FABP1",
        "CA1",
        "EPCAM",
        "RBP2",
        "SLC26A3",
        "CCL20",
        "TNFRSF11A",
        "BEST4",
        "TFF1",
        "MUC2",
        "HCK",
        "TRPM5",
        "SCGN",
        "CHGA",
        "CCL7",
        "CCL13",
    ],
    "Myeloid": [
        "S100B",
        "CD68",
        "XCR1",
        "CLEC9A",
        "CD1C",
        "S100A8",
        "S100A9",
        "TPSAB1",
        "OSM",
    ],
    "T": ["NKG7", "KLRC1", "CCR7", "FOXP3", "CTLA4", "CD8B", "CXCR6", "CD3D"],
    "B": ["MZB1", "IGHA1", "SELL", "CD19", "AICDA"],
    "Stroma": [
        "RSPO3",
        "CCL11",
        "WNT5B",
        "BMP4",
        "CHI3L1",
        "TAGLN",
        "ACTA2",
        "WNT2B",
        "SOX17",
        "RBP7",
        "VWA1",
        "PLVAP",
        "CDH5",
        "CD36",
        "MADCAM1",
        "DARC",
        "RGS5",
        "SOX10",
    ],
}

# top fc from human colon cancer atlas
# markers = ['GPX2',  'OLFM4',  'AHCY',  'PERP',  'SPINK1',  'SLC12A2',  'SLC39A4',  'KRT18',  'PCCA',  'HMGA1',  'ASS1',  'ZFAS1',  'SH3YL1',  'TSPAN8',  'HSPD1',  'PPP1R16A',  'KRT19',  'AGR2',  'EPCAM',  'SFN',  'SPINK4',  'GMDS',  'TFF3',  'STARD10',  'GALNT7',  'MUC2',  'AKAP1',  'ATP1B1',  'CANT1',  'TMEM141',  'PAICS',  'GGCT',  'NME1',  'MGST1',  'HSPE1',  'WDR34',  'CSE1L',  'UBE2S',  'CEACAM6',  'MUC13',  'TSPAN1',  'CEACAM5',  'SLC40A1',  'ISG15',  'MDK',  'PLAC8',  'PIGR',  'SMIM22',  'C19orf33',  'IFI27',  'ERO1A',  'CLDN4',  'HOOK2',  'LLGL2',  'TMEM63A',  'WDR60',  'PAXBP1',  'DGAT1',  'FABP1',  'CC2D1A',  'PYGB',  'AHI1',  'VPS13A',  'SLC26A2',  'FCGBP',  'LGALS4',  'TSTA3',  'S100A6',  'ZG16',  'PHGR1',  'USP53',  'FXYD3',  'TRIB1',  'LDLR',  'ANKRD37',  'APOL1',  'IFI6',  'TSPAN3',  'CD55',  'CXCL16',  'QSOX1',  'KRT8',  'GSKIP',  'CYSTM1',  'ANXA4',  'SOX4',  'MARCKSL1',  'MIB2',  'SPINT2',  'SEC11C',  'NDUFB9',  'NTHL1',  'ZKSCAN1']



## apply binning on spot level data, following current standards, for comparison.

import anndata
import pandas as pd
import scanpy as sc


def cluster_raw_spots(sthdata, resolution=0.5):
    """Example:
    -------
    adata_spot = binning.cluster_raw_spots(sthdata, resolution = 0.2)
    fig = sq.pl.spatial_scatter(adata_spot,
                          color='leiden',
                          crop_coord = [(x1,y1,x2,y2)], # those are original full res image's pixel coordinates
                          title='leiden_'+str(resolution)
                         )

    """
    adata_spot = sthdata.adata.copy()
    sc.pp.normalize_total(adata_spot, target_sum=10e4)  # seurat style
    sc.pp.log1p(adata_spot)
    sc.pp.highly_variable_genes(adata_spot, n_top_genes=4000)
    sc.tl.pca(adata_spot)
    sc.pp.neighbors(adata_spot)
    sc.tl.leiden(adata_spot, resolution=resolution)
    return adata_spot


def get_bin_adata(sthdata, nspot=4):
    """Simple binning

    Example:
    -------
    binadata = get_bin_adata(sthdata, nspot=4) # e.g. 2umx2um into 8umx8um

    """
    adata = sthdata.adata.copy()
    # whether each row index is multiplicate of 4 (plus min), and aggregate based on the two number as index
    # nspot= 4 # 8um=4*2um

    ## get the left top bin start, row and col.
    array_row_min, array_col_min = (
        adata.obs["array_row"].min().astype(int),
        adata.obs["array_col"].min().astype(int),
    )
    adata.obs["array_row_step4"] = (
        nspot * (((adata.obs["array_row"] - array_row_min) / nspot).astype(int))
        + array_row_min
    )
    adata.obs["array_col_step4"] = (
        nspot * (((adata.obs["array_col"] - array_col_min) / nspot).astype(int))
        + array_col_min
    )

    # create adata df
    tmp = adata.to_df().copy()
    tmp = tmp.merge(
        adata.obs[["array_row_step4", "array_col_step4"]],
        how="left",
        left_index=True,
        right_index=True,
    )
    bindf = tmp.groupby(["array_row_step4", "array_col_step4"]).sum()

    # put in original barcodes and create obs
    binobs = bindf.index.to_frame()
    barcodes_included = tmp.groupby(
        ["array_row_step4", "array_col_step4"], observed=True
    ).apply(lambda x: x.index.tolist())
    binobs["barcodes_included"] = barcodes_included

    binobs["spot_index"] = (
        binobs["array_row_step4"].astype(str)
        + "_"
        + binobs["array_col_step4"].astype(str)
    )
    binobs = binobs.set_index("spot_index")

    bindf.index = binobs.index
    # create var, recycle
    binvar = adata.var.copy()

    # create binadata
    binadata = anndata.AnnData(bindf, var=binvar, obs=binobs)
    # create spatial, take mean for each bin
    spatial = pd.DataFrame(
        adata.obsm["spatial"], index=adata.obs.index, columns=["x", "y"]
    )
    for i in binobs.index:
        barcodes = binobs.loc[i]["barcodes_included"]
        a = spatial.loc[barcodes]
        xc, yc = a.mean().values
        binobs.loc[i, "x"] = xc
        binobs.loc[i, "y"] = yc

    binadata.obsm["spatial"] = binobs[["x", "y"]].values

    # put back image
    binadata.uns = adata.uns.copy()

    # get counts
    binadata.obs["n_counts"] = binadata.X.sum(axis=1)
    print(binadata)

    """
    # array_row_col_2_barcode_pd.index.names = ['array_row_step4', 'array_col_step4']
    # original location does not exist 1878,2200
    # sthdata.adata.obs[(sthdata.adata.obs['array_row_int']==1878) & (sthdata.adata.obs['array_col_int']==2200)]
    # so if getting the barcode name, no one
    # bindatadf.merge(array_row_col_2_barcode_pd, how='left',left_index=True, right_index=True)
    """
    return binadata


def get_sthd_guided_bin_adata_v1(
    sthdata,
    pred_col="STHD_pred_ct",
    nspot=4,
    remove_ambiguous_spot=True,
    ambiguous_spot_celltypes=["ambiguous"],
    min_nspot_to_aggregate=2,
):
    """Bin aggregating data, guided by STHD predicted cell type identities of spots. E.g. if one bin has 2 types of spots, they will be aggregated separately, and location is determined by taking mean of x or y for the included spot barcodes
    Params
    ----------
    sthdata:
        STHD class, including adata with spatial, coodinate, fullregimg
    pred_col:
        column in STHD.obs marking STHD predicted cell type
    nspot: int
        Number of  spots to square-bin together, along x or along y.
    remove_ambiguous_spot:
        Whether to remove ambiguous spots or not, based on pred_col
    ambiguous_spot_celltypes:
        name of ambiguous cell identify for spots, e.g. ['ambiguous']
    min_nspot_to_aggregate:
        require that at least min_nspot_to_aggregate number of spots are in the bin to get an aggregated count.
    """
    adata = sthdata.adata.copy()
    array_row_min, array_col_min = (
        adata.obs["array_row"].min().astype(int),
        adata.obs["array_col"].min().astype(int),
    )
    adata.obs["array_row_step4"] = (
        nspot * (((adata.obs["array_row"] - array_row_min) / nspot).astype(int))
        + array_row_min
    )
    adata.obs["array_col_step4"] = (
        nspot * (((adata.obs["array_col"] - array_col_min) / nspot).astype(int))
        + array_col_min
    )

    ########## create adata df ##########
    tmp = adata.to_df()
    tmp = tmp.merge(
        adata.obs[["array_row_step4", "array_col_step4", pred_col]],
        how="left",
        left_index=True,
        right_index=True,
    )
    if remove_ambiguous_spot:  # Optional: remove ambiguous
        tmp = tmp.loc[~tmp[pred_col].isin(ambiguous_spot_celltypes)]  # ambiguous
    bindf = tmp.groupby(
        ["array_row_step4", "array_col_step4", pred_col], observed=True
    ).sum()

    ##########  put in original barcodes and create obs ##########
    binobs = bindf.index.to_frame()
    barcodes_included = tmp.groupby(
        ["array_row_step4", "array_col_step4", pred_col], observed=True
    ).apply(lambda x: x.index.tolist())
    binobs["barcodes_included"] = barcodes_included

    binobs["spot_index"] = (
        binobs["array_row_step4"].astype(str)
        + "_"
        + binobs["array_col_step4"].astype(str)
        + "_"
        + binobs[pred_col].astype(str)
    )
    binobs = binobs.set_index("spot_index")

    bindf.index = binobs.index
    ##########  create var, recycle ##########
    binvar = adata.var.copy()

    ###########  create binadata ##########
    binadata = anndata.AnnData(bindf, var=binvar, obs=binobs)
    ###########  create spatial, take mean for each bin ##########
    spatial = pd.DataFrame(
        adata.obsm["spatial"], index=adata.obs.index, columns=["x", "y"]
    )
    for i in binobs.index:
        barcodes = binobs.loc[i]["barcodes_included"]

        a = spatial.loc[barcodes]
        xc, yc = a.mean().values
        binobs.loc[i, "x"] = xc
        binobs.loc[i, "y"] = yc

    binadata.obsm["spatial"] = binobs[["x", "y"]].values

    ###########  put back image ##########
    binadata.uns = adata.uns.copy()

    ###########  get counts ##########
    binadata.obs["n_counts"] = binadata.X.sum(axis=1)
    binadata.obs["bin_n_spot"] = binadata.obs["barcodes_included"].apply(
        lambda t: len(t)
    )
    # filter minimum spot number
    binadata = binadata[binadata.obs["bin_n_spot"] >= min_nspot_to_aggregate]

    # change list to comma
    binadata.obs["barcodes_included"] = binadata.obs["barcodes_included"].str.join(
        sep=","
    )

    print(binadata)
    return binadata


def remove_celltype_sthd_guided_bin_adata(
    binadata,
    pred_col="STHD_pred_ct",
    remove_ambiguous_spot=True,
    ambiguous_spot_celltypes=["ambiguous"],
):
    """Remove aggregated spots with ambiguous celll types from bindata"""
    if remove_ambiguous_spot:  # Optional: remove ambiguous
        mask = binadata.obs.loc[
            ~binadata.obs[pred_col].isin(ambiguous_spot_celltypes)
        ]  # ambiguous
        binadata = binadata[mask]
    return binadata


def cluster_bin_data(binadata, resolution=1):
    """# cluster bined data

    Example:
    -------
    adata_bin, fig = cluster_bin_data(binadata, coord, resolution=1)
    sc.pl.umap( binadata,        color="leiden")
    fig = sq.pl.spatial_scatter(binadata,
                          color='leiden',
                          crop_coord = coord,
                          title='leiden_'+str(resolution)
                         )

    """
    if ("counts") not in binadata.layers.keys():
        print('[Log] Copying raw counts to adata.layers["counts"]')
        binadata.layers["counts"] = binadata.X.copy()
    print("[Log] Normalizing")
    sc.pp.normalize_total(binadata, target_sum=10e4)
    sc.pp.log1p(binadata)
    sc.pp.highly_variable_genes(binadata, n_top_genes=4000)
    sc.tl.pca(binadata)
    sc.pp.neighbors(binadata)
    print("[Log] umap for visualization")
    sc.tl.umap(binadata)
    print("[Log] leiden clustering")
    sc.tl.leiden(binadata, resolution=resolution)
    return binadata



import matplotlib.pyplot as plt
import squidpy as sq


def spatial_scatter_all_prob(
    sthdata,
    ctcols=None,
    n_plot_cols=6,
    pred_ct_only=True,
    outfile="crop_all_probabilities.png",
    min_num=3,
    single_plot_height=5,
    fig_width=30,
    max_fig_height=60,
):
    """Plot Posterior Probabilities for all STHD pred channels.

    Example:
    -------
    viz.spatial_scatter_all_prob(sthdata,
                         ctcols = config.crc98_order_col,
                         pred_ct_only=False,
                         n_plot_cols = 8,
                         outfile='../crop10_all_probabilities.png'
                        )

    """
    x1, y1, x2, y2 = sthdata.get_sequencing_data_region()
    _ = sthdata.adata.copy()
    if pred_ct_only:
        ctcols = [
            t
            for t in ctcols
            if t
            in (
                set(sthdata.adata.obs["STHD_pred_ct"].values)
                - set(["ambiguous", "filtered"])
            )
        ]
        print(f"[Log] {len(ctcols)} cell types exist in prediction")
        if min_num > 0:
            cts = sthdata.adata.obs.groupby("STHD_pred_ct").count()
            cts = cts[cts[cts.columns[0]] >= min_num]
            ctcols = [t for t in ctcols if t in cts.index.tolist()]
            print(f"[Log] {len(ctcols)} cell types with min number {min_num}")

    if len(ctcols) < n_plot_cols:
        print("[Warning] plotting: select less column. using n ctcols.")
        n_plot_cols = len(ctcols)
    n_row = len(ctcols) // n_plot_cols + 1
    n_col = n_plot_cols
    f, ax = plt.subplots(
        n_row,
        n_col,
        figsize=(fig_width, min(max_fig_height, single_plot_height * n_row)),
    )
    for cur_row in range(n_row):
        for cur_col in range(n_col):
            ind = cur_row * n_col + cur_col
            if ind >= len(ctcols):
                break
            cur_ct = ctcols[ind]
            if cur_ct not in sthdata.adata.obs.columns:
                continue
            sq.pl.spatial_scatter(
                sthdata.adata,
                color=cur_ct,
                crop_coord=[(x1, y1, x2, y2)],
                ax=ax[cur_row, cur_col],
                vmin=0,
                vmax=1,
                colorbar=False,
                legend_fontsize="xx-small",
            )
    plt.savefig(outfile, dpi=300)



"""
Example
python3 STHD/train.py --refile ../testdata/crc_average_expr_genenorm_lambda_98ct_4618gs.txt --patch_list ../testdata/crop10large//patches/52979_9480 ../testdata/crop10large//patches/57479_9480 ../testdata/crop10large//patches/52979_7980 ../testdata/crop10large//patches/55979_7980 ../testdata/crop10large//patches/57479_7980 ../testdata/crop10large//patches/54479_9480 ../testdata/crop10large//patches/55979_9480 ../testdata/crop10large//patches/54479_7980
"""

import argparse
import os
from time import time

import pandas as pd
from STHD import model, qcmask, refscrna, sthdio


def sthdata_match_refgene(sthd_data, refile, ref_gene_filter=True, ref_renorm=False):
    genemeanpd_filtered = refscrna.load_scrna_ref(refile)
    ng1 = sthd_data.adata.shape[1]
    sthd_data.match_refscrna(genemeanpd_filtered, cutgene=True, ref_renorm=ref_renorm)
    ng2 = sthd_data.adata.shape[1]
    print(
        f"cut {ng1} genes to match to reference {ng2} genes",
    )
    if ref_gene_filter:
        genemeanpd_filtered = genemeanpd_filtered.loc[sthd_data.adata.var_names]
    return (sthd_data, genemeanpd_filtered)


def train(sthd_data, n_iter, step_size, beta, debug=False, early_stop=False):
    print("[Log] prepare_constants and training weights")
    X, Y, Z, F, Acsr_row, Acsr_col = model.prepare_constants(sthd_data)
    W, eW, P, Phi, ll_wat, ce_wat, m, v = model.prepare_training_weights(X, Y, Z)
    print("[Log] Training...")
    metrics = model.train(
        n_iter=n_iter,
        step_size=step_size,  # learnin rate
        beta=beta,  # weight of CE w.r.t log-likelihood
        constants=(X, Y, Z, F, Acsr_row, Acsr_col),
        weights=(W, eW, P, Phi, ll_wat, ce_wat, m, v),
        early_stop=early_stop,  # True will trigger early_stop criteria, False will run through all n_iter
    )
    if debug:
        return P, metrics
    else:
        return P


def fill_p_filtered_to_p_full(
    P_filtered, sthd_data_filtered, genemeanpd_filtered, sthd_data
):
    """For prediction performed for filtered data, put P_filtered back to full data size and fill -1.
    sthd_data cannot already have the probability columns
    """
    P_filtered_df = pd.DataFrame(
        P_filtered,
        index=sthd_data_filtered.adata.obs.index,
        columns=genemeanpd_filtered.columns,
    )
    P_filtered_df.columns = "p_ct_" + P_filtered_df.columns
    p_columns = P_filtered_df.columns

    obs_withp = sthd_data.adata.obs.merge(
        P_filtered_df, how="left", left_index=True, right_index=True
    )
    P = obs_withp[list(p_columns)].fillna(-1).values
    return P


def predict(sthd_data, p, genemeanpd_filtered, mapcut=0.8):
    """Based on per barcode per cell type probability, predict cell type and put prediction in adata.
    sthd_data = predict(sthd_data, p, genemeanpd_filtered, mapcut= 0.8)
    """
    adata = sthd_data.adata.copy()
    for i, ct in zip(
        range(len(genemeanpd_filtered.columns)), genemeanpd_filtered.columns
    ):
        adata.obs["p_ct_" + ct] = p[:, i]
    adata.obs["x"] = adata.obsm["spatial"][:, 0]
    adata.obs["y"] = adata.obsm["spatial"][:, 1]

    # get map predictions
    STHD_prob = adata.obs[[t for t in adata.obs.columns if "p_ct_" in t]]
    ct_max = STHD_prob.columns[STHD_prob.values.argmax(1)]
    STHD_pred_ct = pd.DataFrame({"ct_max": ct_max}, index=STHD_prob.index)
    STHD_pred_ct["ct"] = STHD_pred_ct["ct_max"]

    # assign ambiguous based on posterior cut
    ambiguous_mask = (STHD_prob.max(axis=1) < mapcut).values
    STHD_pred_ct.loc[ambiguous_mask, "ct"] = "ambiguous"

    # assign filtered region to 'filtered'
    filtered_mask = (
        STHD_prob.max(1) < 0
    ).values  # by default, filtered spots have prob as -1.
    STHD_pred_ct.loc[filtered_mask, "ct"] = "filtered"

    # assign final cell type prediction
    adata.obs["STHD_pred_ct"] = STHD_pred_ct["ct"]
    print("[Log]Predicted cell type in STHD_pred_ct in adata.obs")
    print(
        "[Log]Predicted cell type probabilities in columns starting with p_ct_ in adata.obs"
    )
    sthd_data.adata = adata

    return sthd_data


########## Training IO: Saving


def save_prediction_pdata(sthdata, file_path=None, prefix=""):
    """Save from sthdata the pdata into dataframe with probabilities, predicted cell type, and x, y

    Example:
    -------
    pdata = train.save_prediction_pdata(sthdata, file_path = '', prefix = '')

    """
    predcols = (
        ["x", "y"]
        + ["STHD_pred_ct"]
        + [t for t in sthdata.adata.obs.columns if "p_ct_" in t]
    )
    pdata = sthdata.adata.obs[predcols]

    if file_path is not None:
        pdata_path = os.path.join(file_path, prefix + "_pdata.tsv")
        pdata.to_csv(pdata_path, sep="\t")
        print(f"[Log] prediction saved to {pdata_path}")
    return pdata


########## Training IO: Loading


def load_data(file_path):
    """Load expr data. Only works with cropped data."""
    sthd_data = sthdio.STHD(
        spatial_path=os.path.join(file_path, "adata.h5ad.gzip"),
        counts_data=None,
        full_res_image_path=os.path.join(file_path, "fullresimg_path.json"),
        load_type="crop",
    )
    print("[log] Number of spots: ", sthd_data.adata.shape[0])
    return sthd_data


def load_pdata(file_path, prefix=""):
    pdata_path = os.path.join(file_path, prefix + "_pdata.tsv")
    pdata = pd.read_table(pdata_path, index_col=0)
    return pdata


def add_pdata(sthd_data, pdata):
    """Load from pdata into dataframe with probabilities, predicted cell type, and x, y, and put in sthdata
    to rename: put_pdata_to_sthdata

    Example:
    -------
    pdata = train.add_pdata(sthdata, pdata)

    """
    sthdata = sthd_data
    exist_cols = sthdata.adata.obs.columns.intersection(pdata.columns)
    print("[Log] Loading prediction into sthdata.adata.obs, overwriting")
    print(exist_cols)
    for col in sthdata.adata.obs[exist_cols]:
        del sthdata.adata.obs[col]
    sthdata.adata.obs = sthdata.adata.obs.merge(
        pdata, how="left", left_index=True, right_index=True
    )
    return sthdata


def load_data_with_pdata(file_path, pdata_prefix=""):
    """A simplified
    Load full prediction for the patch. need os.path.join(file_path, pdata_prefix+_pdata.h5ad)
    Return adata with probability and predicted cell type in .obs
    contains all gene expression
    [] todo: deprecate
    """
    sthdata = load_data(file_path)
    pdata = load_pdata(file_path, pdata_prefix)
    sthdata_with_pdata = add_pdata(sthdata, pdata)
    return sthdata_with_pdata


##########


def main(args):
    start = time()
    for patch_path in args.patch_list:
        print(f"[log] {time() - start:.2f}, start processing patch {patch_path}")
        if args.filtermask:
            sthdata = load_data(patch_path)
            print(sthdata.adata.shape)
            sthdata.adata = qcmask.background_detector(
                sthdata.adata,
                threshold=args.filtermask_threshold,
                n_neighs=4,
                n_rings=args.filtermask_nrings,
            )
            print(sthdata.adata.shape)
            # visualize_background(sthdata)
            sthdata_filtered = qcmask.filter_background(
                sthdata, threshold=args.filtermask_threshold
            )
            sthdata_filtered, genemeanpd_filtered = sthdata_match_refgene(
                sthdata_filtered, args.refile, ref_renorm=args.ref_renorm
            )
            print("[Log]Training")
            P_filtered = train(sthdata_filtered, args.n_iter, args.step_size, args.beta)
            P = fill_p_filtered_to_p_full(
                P_filtered, sthdata_filtered, genemeanpd_filtered, sthdata
            )
        else:
            sthdata = load_data(patch_path)
            sthdata, genemeanpd_filtered = sthdata_match_refgene(
                sthdata, args.refile, ref_renorm=args.ref_renorm
            )
            print("[Log]Training")
            P = train(sthdata, args.n_iter, args.step_size, args.beta)

        sthdata = predict(sthdata, P, genemeanpd_filtered, mapcut=args.mapcut)
        _ = save_prediction_pdata(sthdata, file_path=patch_path, prefix="")
        print("[Log]prediction saved")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--n_iter",
        default=23,
        type=int,
        help="iteration to optimize LL and CE. recommended 23 (tuned for human colon cancer sample)0",
    )
    parser.add_argument("--step_size", default=1, type=int)
    parser.add_argument(
        "--beta",
        default=0.1,
        type=float,
        help="beta parameter for borrowing neighbor info. recommended 0.1",
    )
    parser.add_argument(
        "--mapcut",
        default=0.8,
        type=float,
        help="posterior cutoff for celltype prediction",
    )
    parser.add_argument(
        "--refile", type=str, help="reference normalized gene mean expression."
    )
    parser.add_argument(
        "--ref_renorm", default=False, type=bool, help="recommended False"
    )
    parser.add_argument(
        "--filtermask", type=bool, default=True, help="whether to filter masked spots"
    )
    parser.add_argument(
        "--filtermask_nrings",
        default=2,
        type=int,
        help="auto detection of low-count region: number of rings to consider neighbor",
    )
    parser.add_argument(
        "--filtermask_threshold",
        default=51,
        type=int,
        help="auto detection of low-count region: number of total counts",
    )
    parser.add_argument(
        "--patch_list", nargs="+", default=[], help="a space separated patch path list"
    )

    args = parser.parse_args()
    
    main(args)
    
    """
    # quick test
    class Args:
    def __init__(self):
        self.n_iter=10
        self.step_size = 1
        self.beta = 0.1
        self.refile =  '../testdata/crc_average_expr_genenorm_lambda_98ct_4618gs.txt'
        self.filtermask = True
        self.patch_list = ['../testdata/crop10']
    train.main(Args())
    """



import numpy as np
import squidpy as sq
from numba import njit, prange

"""library to train STHD model. Example codes:
import pandas as pd
import matplotlib.pyplot as plt
import squidpy as sq
import sys
sys.path.append('../STHD')
import model
import refscrna
import sthdio

# load data
file_path = '../testdata/crop10'
sthd_data = sthdio.STHD(
            spatial_path = f'{file_path}/adata.h5ad',
            counts_data = None, 
            full_res_img = f'{file_path}/fullresimg.npy', 
            full_res_img_coor = f'{file_path}/coor.json', 
            load_type = 'crop'
        )
print('Number of spots: ', sthd_data.adata.shape[0])
refile = '../testdata/crc_average_expr_genenorm_lambda_98ct_4618gs.txt'
genemeanpd_filtered = refscrna.load_scrna_ref(refile)
sthd_data.match_refscrna(genemeanpd_filtered, cutgene=True)

# prepare_constants and training weights
X, Y, Z, F, Acsr_row, Acsr_col = model.prepare_constants(sthd_data)
W, eW, P, Phi, ll_wat, ce_wat, m, v = model.prepare_training_weights(X,Y,Z)

# train
metrics = model.train(
    n_iter = 20, 
    step_size = 1, # learnin rate
    beta = 0.1,  # weight of CE w.r.t log-likelihood
    constants = (X, Y, Z, F, Acsr_row, Acsr_col), 
    weights = (W, eW, P, Phi, ll_wat, ce_wat, m, v)
)

# visualize results
color_list1 = ['STHD_pred_ct', 'p_ct_cP2 (Plasma IgG)', 'p_ct_cP3 (Plasma IgG prolif)']
color_list2 = ['p_ct_Tumor cE05 (Enterocyte 2)', 'p_ct_cS04 (Endo)', 'p_ct_Tumor cE11 (Enteroendocrine)']
model.analysis(sthd_data, P, genemeanpd_filtered, color_list1, color_list2, mapcut = 0.8)
"""


@njit(parallel=True, fastmath=True)
def fill_F(X, Y, Z, N, Lambda, D, F):
    for a in prange(X):
        for t in range(Z):
            for g in range(Y):
                tmp = D[a] * Lambda[t, g]
                F[a, t] = F[a, t] + (N[a, g] * np.log(tmp) - tmp)


def prepare_constants(sthd_data):
    X = sthd_data.adata.obs.shape[0]  # n of spot
    Y = sthd_data.adata.shape[1]  # n of gene (filtered )
    Z = sthd_data.lambda_cell_type_by_gene_matrix.shape[0]  # n of cell type

    # get raw data
    N = sthd_data.adata.to_df().values  # [X,Y] (a,g) number of total reads in each spot
    Lambda = sthd_data.lambda_cell_type_by_gene_matrix.astype(
        "float32"
    )  # [Z,Y] (t,g) gene's relative expression in each cell type.
    D = (
        np.squeeze(np.asarray(sthd_data.adata.X.sum(axis=1))) + 0.1
    )  # [X] (a) spot depth. adding 0.1 to prevent explod

    # get neighbor adjacency matrix
    sq.gr.spatial_neighbors(
        sthd_data.adata, spatial_key="spatial", coord_type="grid", n_neighs=4, n_rings=2
    )
    A_csr = sthd_data.adata.obsp["spatial_connectivities"]  # [X, X]
    print("Currently we only support symmetric adjacency matrix of neighbors")
    Acsr_row = A_csr.indptr
    Acsr_col = A_csr.indices

    F = np.zeros([X, Z], dtype="float32")
    fill_F(X, Y, Z, N, Lambda, D, F)
    return X, Y, Z, F, Acsr_row, Acsr_col


def prepare_training_weights(X, Y, Z):
    # prepare training parameters
    W = np.ones([X, Z]).astype("float32")  # initialization
    # prepare derived parameters
    eW = np.zeros([X, Z], dtype="float32")
    P = np.zeros([X, Z], dtype="float32")
    Phi = np.zeros([X], dtype="float32")
    # prepare deriatives
    ll_wat = np.zeros([X, Z], dtype="float32")
    ce_wat = np.zeros([X, Z], dtype="float32")
    # additional variables for adam
    m = np.zeros([X, Z], dtype="float32")
    v = np.zeros([X, Z], dtype="float32")
    return W, eW, P, Phi, ll_wat, ce_wat, m, v


def early_stop_criteria(metrics, beta, n=5, threshold=0.01):
    if len(metrics) < n:
        return False
    else:
        metrics_check = metrics[-n:]
        ll = [i[0] for i in metrics_check]
        ll_range = max(ll) - min(ll)
        ll_ratio = ll_range / np.abs(ll[-1])
        if beta == 0:
            # in this case we won't check ce for early stopping
            ce_ratio = 0
        else:
            ce = [i[1] for i in metrics_check]
            ce_range = max(ce) - min(ce)
            ce_ratio = ce_range / np.abs(ce[-1])
        if (ll_ratio < threshold) and (ce_ratio < threshold):
            return True
        else:
            return False


def early_stop_criteria_2(metrics, beta, n=10, threshold=0.01):
    if len(metrics) < n:
        return False
    else:
        metrics_check = metrics[-n:]
        loss = [-i[0] + beta * i[1] for i in metrics_check]
        loss_range = max(loss) - min(loss)
        loss_ratio = loss_range / np.abs(loss[-1])
        if loss_ratio < threshold:
            return True
        else:
            return False


def train(
    n_iter,
    step_size,
    beta,
    constants,
    weights,
    early_stop=False,
    beta1=0.9,
    beta2=0.999,
    epsilon=1e-8,
):
    X, Y, Z, F, Acsr_row, Acsr_col = constants
    W, eW, P, Phi, ll_wat, ce_wat, m, v = weights

    metrics = []
    for i in range(n_iter):
        update_eW(eW, W, X, Y, Z)
        update_Phi(Phi, eW, X, Y, Z)
        update_P(P, eW, Phi, X, Y, Z)
        update_ll_wat(ll_wat, P, F, X, Y, Z)
        update_ce_wat(ce_wat, P, Acsr_row, Acsr_col, X, Y, Z)
        update_m_v(m, v, beta1, beta2, beta, ll_wat, ce_wat, X, Y, Z)
        update_W_adam(W, m, v, beta1, beta2, i + 1, step_size, epsilon, X, Y, Z)
        ll = calculate_ll(P, F, X, Y, Z)
        ce = calculate_ce(P, Acsr_row, Acsr_col, X, Y, Z)
        metrics.append((ll, ce))
        print(i, -ll + beta * ce, ll, ce)

        if early_stop and early_stop_criteria_2(metrics, beta):
            return metrics

    return metrics


# ----------------------- update trainable parameters -------------------
@njit(parallel=True, fastmath=True)
def update_eW(eW, W, X, Y, Z):
    # inplace update eW using W
    for a in prange(X):
        for t in range(Z):
            eW[a, t] = np.exp(W[a, t])


@njit(parallel=True, fastmath=True)
def update_Phi(Phi, eW, X, Y, Z):
    # inplace update Phi using eW
    for a in prange(X):
        Phi[a] = 0
        for t in range(Z):
            Phi[a] = Phi[a] + eW[a, t]


@njit(parallel=True, fastmath=True)
def update_P(P, eW, Phi, X, Y, Z):
    # inplace update P using eW, Phi
    for a in prange(X):
        for t in range(Z):
            P[a, t] = eW[a, t] / Phi[a]


# ----------------------- calculate losses -------------------
@njit
def csr_obtain_column_index_for_row(row, column, i):
    # get row i's non-zero items' column index
    row_start = row[i]
    row_end = row[i + 1]
    column_indices = column[row_start:row_end]
    return column_indices


@njit
def csr_obtain_value_by_i_j(row, column, value, i, j):
    # get row i's non-zero items' column index
    row_start = row[i]
    row_end = row[i + 1]
    column_indices = column[row_start:row_end]
    row_values = value[row_start:row_end]
    if j in set(column_indices):
        for k, cid in enumerate(column_indices):
            if cid == j:
                v = row_values[k]
                break
    else:
        v = 0
    return v


@njit
def calculate_ce(P, Acsr_row, Acsr_col, X, Y, Z):
    # ! DO NOT use numba parallel in this function!!
    res = 0
    for a in range(X):
        neighbors = csr_obtain_column_index_for_row(Acsr_row, Acsr_col, a)
        for t in range(Z):
            cur = 0
            for a_star in neighbors:
                cur = cur + np.log(P[a_star, t])
            res = res - P[a, t] * cur
    res = res / X
    return res


@njit(parallel=True, fastmath=True)
def calculate_ll(P, F, X, Y, Z):
    res = 0
    for a in prange(X):
        for t in range(Z):
            res = res + P[a, t] * F[a, t]
    res = res / X
    return res


# ----------------------- calculate gradients -------------------
@njit(parallel=True, fastmath=True)
def update_ll_wat(ll_wat, P, F, X, Y, Z):
    for a_tilda in prange(X):
        for t_tilda in range(Z):
            cur = 0
            for t in range(Z):
                cur = cur + P[a_tilda, t] * P[a_tilda, t_tilda] * F[a_tilda, t]
            ll_wat[a_tilda, t_tilda] = (
                -1 * cur + P[a_tilda, t_tilda] * F[a_tilda, t_tilda]
            )
            ll_wat[a_tilda, t_tilda] = ll_wat[a_tilda, t_tilda] / X


@njit(parallel=True, fastmath=True)
def update_ce_wat(ce_wat, P, Acsr_row, Acsr_col, X, Y, Z):
    for a_tilda in prange(X):
        neighbors = csr_obtain_column_index_for_row(Acsr_row, Acsr_col, a_tilda)
        # Since Acsr is symetric, neighbors of a_tilda is also the spots whose neighbor contains a_tilda
        for t_tilda in range(Z):
            comp1 = 0  # neighbors of a_tilda
            comp2 = 0  # a_tilda is other's neighbo
            for a_star in neighbors:
                cur1 = 0
                for t in range(Z):
                    cur1 = cur1 + P[a_tilda, t] * np.log(P[a_star, t])
                comp1 = comp1 + np.log(P[a_star, t_tilda]) - cur1
                comp2 = comp2 + P[a_star, t_tilda] - P[a_tilda, t_tilda]
            ce_wat[a_tilda, t_tilda] = (-P[a_tilda, t_tilda] * comp1 - comp2) / X


# ----------------------- (DEPRECATED! USE ADAM!) Optimizer: gradient descent -------------------
@njit(parallel=True, fastmath=True)
def update_W_gd(W, ll_wat, ce_wat, alpha, beta, X, Y, Z):
    for a in prange(X):
        for t in range(Z):
            W[a, t] = W[a, t] - alpha * (-ll_wat[a, t] + beta * ce_wat[a, t])


# ----------------------- Optimizer: ADAM -------------------
# following implementation in https://pytorch.org/docs/stable/generated/torch.optim.Adam.html
# paper: https://arxiv.org/pdf/1412.6980.pdf
@njit(parallel=True, fastmath=True)
def update_m_v(m, v, beta1, beta2, beta, ll_wat, ce_wat, X, Y, Z):
    for a in prange(X):
        for t in range(Z):
            cur_gredient = -ll_wat[a, t] + beta * ce_wat[a, t]
            m[a, t] = m[a, t] * beta1 + (1 - beta1) * cur_gredient
            v[a, t] = v[a, t] * beta2 + (1 - beta2) * cur_gredient**2


@njit(parallel=True, fastmath=True)
def update_W_adam(W, m, v, beta1, beta2, i, alpha, epsilon, X, Y, Z):
    for a in prange(X):
        for t in range(Z):
            m_correct = m[a, t] / (1 - beta1**i)
            v_correct = v[a, t] / (1 - beta2**i)
            W[a, t] = W[a, t] - alpha * m_correct / (v_correct**0.5 + epsilon)



"""Split data into patches for prediction, or combine patched predictions into a single one.

# Spliting patches from a test large cropped data:
python3 patchify.py \
    --spatial_path ../testdata/crop10_large/adata.h5ad.gzip \
    --full_res_image_path ../testdata/crop10_large/fullresimg_path.json \
    --load_type crop \
    --dx 1500 \
    --dy 1500 \
    --scale_factor 0.07973422 \
    --refile ../testdata/crc_average_expr_genenorm_lambda_98ct_4618gs.txt \
    --save_path ../testdata/crop10_large_patchify \
    --mode split

# Spliting patches from the full-size original data:
python3 patchify.py \
    --spatial_path ../testdata/VisiumHD/square_002um/ \
    --counts_data filtered_feature_bc_matrix.h5 \
    --full_res_image_path ../testdata/VisiumHD/Visium_HD_Human_Colon_Cancer_tissue_image.btf \
    --load_type original \
    --dx 1500 \
    --dy 1500 \
    --scale_factor 0.07973422 \
    --refile ../testdata/crc_average_expr_genenorm_lambda_98ct_4618gs.txt \
    --save_path ../analysis/full_patchify \
    --mode split

# Combine predictions
python3 patchify.py \
    --refile ../testdata/crc_average_expr_genenorm_lambda_98ct_4618gs.txt \
    --save_path ../testdata/crop10_large_patchify \
    --mode combine
"""

import argparse
import os
import pathlib
from collections import defaultdict

import numpy as np
import pandas as pd
from tqdm import tqdm
from STHD import sthdio, train


def _load_into_dict(res_dict, file, columns):
    data = train.load_pdata(file)
    indices = data.index.tolist()
    cur_columns = data.columns.tolist()
    if columns != cur_columns:
        raise ValueError("Patches have mismatched column names")
    if columns[:3] != ["x", "y", "STHD_pred_ct"]:
        raise ValueError("Patch column orders are not correct")
    values = data.values

    for i, barcode in enumerate(indices):
        res_dict[barcode].append(values[i])


def _process_barcode(res_dict, columns):
    """Implement the following logic:

    The input res_dict already groups data by barcode. For each barcode, it may contain one or more records.
    In this process, we will recalculate each barcode's probability predictions:
        1. IF this barcode only has one row: directly use that row's original p predictions
        2. IF all rows are filtered -> directly use arbitrary row's original p predictions
        3. IF not all rows are filtered:
           1. ignore filtered rows
           2. for remaining rows, average probabilities
    """
    id_STHD_pred_ct = columns.index("STHD_pred_ct")

    for barcode in tqdm(res_dict):
        data = np.array(res_dict[barcode])
        data = np.delete(data, id_STHD_pred_ct, axis=1)
        data_non_filtered = data[data[:, -1] != -1]

        if len(data) == 1:
            res_dict[barcode] = data[0]
        elif len(data_non_filtered) == 0:
            res_dict[barcode] = data[0]
        else:
            res_dict[barcode] = data_non_filtered.mean(axis=0)


def _combine_patch(patch_dir):
    files = [os.path.join(patch_dir, f) for f in os.listdir(patch_dir)]
    res_dict = defaultdict(list)
    columns = train.load_pdata(files[0]).columns.tolist()

    print("[log] Loading patches")
    for file in tqdm(files):
        _load_into_dict(res_dict, file, columns)

    print("[log] Process probabilities in each barcodes")
    _process_barcode(res_dict, columns)

    print("[log] Reshape data into pandas dataframe")
    columns_remove_prediction = columns.copy()
    columns_remove_prediction.remove("STHD_pred_ct")
    pdata = pd.DataFrame.from_dict(
        res_dict, orient="index", columns=columns_remove_prediction
    )

    return pdata


def _patchify(x1, y1, x2, y2, w, h, dw=10, dh=10):
    # (x1, y1, x2, y2): coordinate of the region to patchify
    # w: width of each path. x1, x2 is in width direction.
    # h: height of each path. y1, y2 is in height direction.
    # dw: overlap pixels in width direction
    # dh: overlap pixels in height direction
    print("Caution!! Patchify should be ran on non-cropped data only!")
    xs = list(range(x1, x2, w)) + [x2]
    ys = list(range(y1, y2, h)) + [y2]
    patches = []
    for i in range(len(xs) - 1):
        for j in range(len(ys) - 1):
            patches.append([xs[i] - dw, ys[j] - dh, xs[i + 1] + dw, ys[j + 1] + dh])
    return patches


def patchify(sthd_data, save_path, x1, y1, x2, y2, dx, dy, scale_factor):
    allregion_path = f"{save_path}/all_region"
    patch_path = f"{save_path}/patches"

    pathlib.Path(allregion_path).mkdir(parents=True, exist_ok=True)
    pathlib.Path(patch_path).mkdir(parents=True, exist_ok=True)

    all_region = sthd_data.crop(
        x1,
        x2,
        y1,
        y2,
        scale_factor,
    )
    all_region.save(allregion_path)

    patches = _patchify(x1, y1, x2, y2, dx, dy)
    # cautious! The patch must contains sequensing data, otherwise it will return error.
    for patch in patches:
        x1_patch, y1_patch, x2_patch, y2_patch = patch
        crop_data = sthd_data.crop(
            x1_patch,
            x2_patch,
            y1_patch,
            y2_patch,
            scale_factor,
        )
        crop_data.save(f"{patch_path}/{x1_patch}_{y1_patch}")


def merge(save_path, refile):
    allregion_path = f"{save_path}/all_region"
    patch_path = f"{save_path}/patches"

    # load sthd_data that were patchified
    sthdata = train.load_data(allregion_path)
    sthdata, genemeanpd_filtered = train.sthdata_match_refgene(
        sthdata, refile, ref_gene_filter=True, ref_renorm=False
    )
    pdata = _combine_patch(patch_path)

    # Align the pdata's barcode order to sthdata
    align_sthdata_pdata = train.add_pdata(sthdata, pdata)
    pdata_reorder = align_sthdata_pdata.adata.obs[
        [t for t in align_sthdata_pdata.adata.obs.columns if "p_ct_" in t]
    ]
    if [
        "p_ct_" + i for i in genemeanpd_filtered.columns.tolist()
    ] != pdata_reorder.columns.tolist():
        raise ValueError("Cell type order miss aligned")

    # Predict cell types and save results
    sthdata_with_pdata = train.predict(
        sthdata, pdata_reorder.values, genemeanpd_filtered, mapcut=0.8
    )
    _ = train.save_prediction_pdata(sthdata_with_pdata, file_path=allregion_path)


def main(args):
    if args.mode == "combine":
        merge(args.save_path, args.refile)
    elif args.mode == "split":
        # load data
        full_data = sthdio.STHD(
            spatial_path=args.spatial_path,
            counts_data=args.counts_data,
            full_res_image_path=args.full_res_image_path,
            load_type=args.load_type,
        )

        # get boundrary of the input file
        x1, y1, x2, y2 = full_data.get_sequencing_data_region()
        x1 = max(0, x1)
        y1 = max(0, y1)
        x2 = max(0, x2)
        y2 = max(0, y2)

        # split input into patches
        patchify(
            full_data,
            args.save_path,
            x1,
            y1,
            x2,
            y2,
            args.dx,
            args.dy,
            args.scale_factor,
        )

        # generate command for training through patches
        patch_dir = os.path.join(args.save_path, "patches")
        files = [os.path.join(patch_dir, f) for f in os.listdir(patch_dir)]
        cmd = f"""python3 -m STHD.train --refile {args.refile} --patch_list {" ".join(files)}"""
        print("")
        print("Please copy and run the following command to train through patches:")
        print(cmd)
    else:
        raise ValueError("Unexpected args.mode")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--spatial_path", type=str)
    parser.add_argument("--counts_data", default="", type=str)
    parser.add_argument("--full_res_image_path", type=str)
    parser.add_argument("--load_type", type=str)
    parser.add_argument("--dx", type=int)
    parser.add_argument("--dy", type=int)
    parser.add_argument("--scale_factor", type=float)
    parser.add_argument("--refile", type=str, required=True)
    parser.add_argument("--save_path", type=str, required=True)
    parser.add_argument("--mode", type=str, required=True)
    args = parser.parse_args()

    main(args)



"""Plot Results as Interactive HTML via The STHD Visualization Module.

Usage:
python3 sthdviz.py \
    --patch_path ../analysis/full_patchify \
    --title full_patchify
"""

import argparse

import numpy as np
from bokeh.layouts import column
from bokeh.models import (
    BoxZoomTool,
    CategoricalColorMapper,
    ColorBar,
    CustomJS,
    CustomJSHover,
    CustomJSTickFormatter,
    FixedTicker,
    HoverTool,
    LinearColorMapper,
    LogColorMapper,
    MultiChoice,
    WheelZoomTool,
)
from bokeh.models.labeling import AllLabels
from bokeh.plotting import figure, output_file, show
from tqdm import tqdm

from STHD import color_palette, train


def rasterize(df, val_col):
    """Rasterize the input dataframe with categorical values
    df: input dataframe to rasterize
        array_row: index for row
        array_col: index for col
        {val_col}: the cell value (usually it is cell type prediction label)
    """
    df = df.dropna()
    row_min, row_max = df["array_row"].min(), df["array_row"].max()
    col_min, col_max = df["array_col"].min(), df["array_col"].max()
    val_char_len = df[val_col].str.len().max()
    print(row_min, row_max, len(set(df["array_row"].values)))
    print(col_min, col_max, len(set(df["array_col"].values)))

    row_ids = (df["array_row"].values - row_min).astype(int)
    col_ids = (df["array_col"].values - col_min).astype(int)
    vals = df[val_col].values

    nrow = max(row_ids) + 1
    ncol = max(col_ids) + 1

    res = np.array([[""] * ncol] * nrow, dtype=f"<U{val_char_len}")

    for i, val in tqdm(enumerate(vals), total=len(vals)):
        res[row_ids[i]][col_ids[i]] = val
    return res


def rasterize_numerical(df, val_col):
    """Rasterize the input dataframe with numerical values
    df: input dataframe to rasterize
        array_row: index for row
        array_col: index for col
        {val_col}: the cell value (usually it is cell type prediction label)
    """
    df = df.dropna()
    row_min, row_max = df["array_row"].min(), df["array_row"].max()
    col_min, col_max = df["array_col"].min(), df["array_col"].max()
    print(row_min, row_max, len(set(df["array_row"].values)))
    print(col_min, col_max, len(set(df["array_col"].values)))

    row_ids = (df["array_row"].values - row_min).astype(int)
    col_ids = (df["array_col"].values - col_min).astype(int)
    vals = df[val_col].values

    nrow = max(row_ids) + 1
    ncol = max(col_ids) + 1

    res = np.zeros((nrow, ncol), dtype=np.float32)

    for i, val in tqdm(enumerate(vals), total=len(vals)):
        res[row_ids[i]][col_ids[i]] = val
    return res


def convert_numerical(df, cmap):
    """Convert cell types into numerical representations using two digits of hex numbers."""
    if len(cmap) > 254:
        raise ValueError("this function can only process up to 254 cell types")

    hex_dict = dict()
    num_dict = dict()
    for i, cell_type in enumerate(cmap):
        hex_dict[cell_type] = hex(i)[2:]
        num_dict[cell_type] = str(i)
    hex_dict[""] = hex(len(cmap))[2:]
    num_dict[""] = str(len(cmap))

    df_hex = np.empty(df.shape, dtype="<U2")
    for i in tqdm(
        range(df.shape[0]),
        total=len(df),
        desc="convert cell types to numerical representations",
    ):
        for j in range(df.shape[1]):
            df_hex[i][j] = hex_dict[df[i][j]]
    return df_hex, hex_dict, num_dict


def fast_plot(
    rastered_df,
    cmap=color_palette.get_config_colormap(name="colormap_coloncatlas_98"),
    cmap_light=color_palette.get_config_colormap(name="colormap_coloncatlas_98_light"),
    cmap_dark=color_palette.get_config_colormap(name="colormap_coloncatlas_98_dark"),
    title="STHD_visualization",
    square_size=7,
    save_root_dir="",
):
    """Interactive plotting by rasterizing the scatter plot into an image.

    rastered_df: generated by the rasterize function above.
    """
    # --------------------------- Save Location ----------------------------------
    if len(save_root_dir) > 0:
        output_file(filename=f"{save_root_dir}/{title}.html", title=title)

    # ---------------------------Numerical representation of input df------------
    rastered_df_hex, ct_hex_map, ct_num_map = convert_numerical(rastered_df, cmap)
    cmap_hex = {ct_hex_map[cell_type]: cmap[cell_type] for cell_type in cmap}
    cmap_hex_light = {
        ct_hex_map[cell_type]: cmap_light[cell_type] for cell_type in cmap_light
    }
    cmap_hex_dark = {
        ct_hex_map[cell_type]: cmap_dark[cell_type] for cell_type in cmap_dark
    }
    hex_ct_map = {ct_hex_map[ct]: ct for ct in ct_hex_map}
    num_ct_map = {ct_num_map[ct]: ct for ct in ct_num_map}

    # -------------------------- prepare colormap -----------------------------------
    cell_type_hex = list(cmap_hex.keys())
    color_list = list(cmap_hex.values())

    mapper = CategoricalColorMapper(factors=cell_type_hex, palette=color_list)
    color_key_pos = dict(
        zip(cell_type_hex, list(zip(range(len(cmap_hex)), list(color_list))))
    )

    ticker = FixedTicker(ticks=np.arange(len(num_ct_map)))
    formatter = CustomJSTickFormatter(
        args={"num_ct_map": num_ct_map}, code="return num_ct_map[tick]"
    )

    color_bar = ColorBar(
        color_mapper=mapper,
        ticker=ticker,
        formatter=formatter,
        major_label_text_font_size="5px",
        major_label_policy=AllLabels(),
        label_standoff=6,
        border_line_color=None,
    )

    # -------------------------- Prepare figure ------------------------------------
    TOOLS = "save,pan,reset,wheel_zoom"
    p = figure(
        title=title,
        x_axis_location="above",
        frame_width=1700,
        frame_height=800,
        match_aspect=True,
        tools=TOOLS,
        toolbar_location="above",
    )
    p.toolbar.active_scroll = p.select_one(WheelZoomTool)
    p.grid.grid_line_color = None
    p.axis.axis_line_color = None
    p.axis.major_tick_line_color = None
    p.axis.major_label_text_font_size = "7px"
    p.axis.major_label_standoff = 0
    p.title.align = "center"

    p.add_tools(BoxZoomTool(match_aspect=True))
    p.add_tools(
        HoverTool(
            tooltips=[("class", "@image{custom}")],
            formatters={
                "@image": CustomJSHover(
                    code="return hex_ct_map[value];", args={"hex_ct_map": hex_ct_map}
                )
            },
        )
    )

    # ---------------------------Draw image -------------------------------------------
    img = p.image(
        image=[rastered_df_hex],
        x=0,
        y=0,
        dw=rastered_df_hex.shape[1],
        dh=rastered_df_hex.shape[0],
        color_mapper=mapper,
    )

    p.grid.grid_line_color = None
    p.axis.axis_line_color = None
    p.axis.major_tick_line_color = None
    p.axis.major_label_text_font_size = "7px"
    p.axis.major_label_standoff = 0
    p.add_layout(color_bar, "right")
    p.y_range.flipped = False

    # ----------------------Multiple Choice module ------------------------------------------------
    # multi choice
    OPTIONS = sorted(list((cmap.keys())))
    multi_choice = MultiChoice(
        value=[], options=OPTIONS, title="Selection:", min_width=300
    )
    callback = CustomJS(
        args=dict(
            img=img,
            color_key_pos=color_key_pos,
            cmap_light=cmap_hex_light,
            cmap_dark=cmap_hex_dark,
            ct_hex_map=ct_hex_map,
        ),
        code="""
    // reset colors
    const length = this.value.length;
    console.log(length);
    if (length===0) { // if no cell type is selected, use default colors
        for (const ct_hex in color_key_pos) {
            img.glyph.color_mapper.palette[color_key_pos[ct_hex][0]] = color_key_pos[ct_hex][1];
        }
    } else { // if there are cell type selected, set all cell types to have ligth colors (and later highlight selected cell types)
        for (const ct_hex in color_key_pos) {
            img.glyph.color_mapper.palette[color_key_pos[ct_hex][0]] = cmap_light[ct_hex];
        }
    }

    // highlight selected genes
    for (const ct of this.value) {
        const ct_hex = ct_hex_map[ct];
        img.glyph.color_mapper.palette[color_key_pos[ct_hex][0]] = cmap_dark[ct_hex];
    }
    
    // apply changes and update plot
    img.glyph.change.emit();
    """,
    )
    multi_choice.js_on_change("value", callback)
    show(column(multi_choice, p))


def fast_plot_numerical(
    rastered_df,
    title="STHD_visualization",
    square_size=7,
    save_root_dir="",
    mapper_type="linear",
    palette="Viridis256",
):
    """Interactive plotting by rasterizing the scatter plot into an image.

    rastered_df: generated by the rasterize function above.
    """
    if mapper_type == "log":
        cls = LogColorMapper
        if rastered_df.min() < 0:
            raise ValueError(
                "Cannot us log scale colormap since there are negative values in rastered_df"
            )
        min_val = np.min(
            rastered_df[np.nonzero(rastered_df)]
        )  # get non-zero minimum values
        if min_val > 1:
            rastered_df += 1  # in log scale, the minimum value is still zero
        else:
            rastered_df += min_val * 0.1
    else:
        cls = LinearColorMapper

    mapper = cls(palette=palette, low=rastered_df.min(), high=rastered_df.max())

    if len(save_root_dir) > 0:
        output_file(filename=f"{save_root_dir}/{title}.html", title=title)

    TOOLS = "hover,save,pan,reset,wheel_zoom"
    p = figure(
        title=title,
        x_axis_location="above",
        frame_width=1700,
        frame_height=800,
        match_aspect=True,
        tools=TOOLS,
        toolbar_location="above",
        tooltips=[("x", "$x{0}"), ("y", "$y{0}"), ("class", "@image")],
    )
    p.add_tools(BoxZoomTool(match_aspect=True))
    p.toolbar.active_scroll = p.select_one(WheelZoomTool)
    p.grid.grid_line_color = None
    p.axis.axis_line_color = None
    p.axis.major_tick_line_color = None
    p.axis.major_label_text_font_size = "7px"
    p.axis.major_label_standoff = 0
    p.title.align = "center"

    _ = p.image(
        image=[rastered_df],
        x=0,
        y=0,
        dw=rastered_df.shape[1],
        dh=rastered_df.shape[0],
        color_mapper=mapper,
    )

    p.grid.grid_line_color = None
    p.axis.axis_line_color = None
    p.axis.major_tick_line_color = None
    p.axis.major_label_text_font_size = "7px"
    p.axis.major_label_standoff = 0

    color_bar = ColorBar(
        color_mapper=mapper,
        major_label_text_font_size="3px",
        label_standoff=6,
        border_line_color=None,
    )
    p.add_layout(color_bar, "right")
    p.y_range.flipped = False

    show(p)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--patch_path", type=str, required=True)
    parser.add_argument("--title", default="sthd_pred_cell_type", type=str)
    args = parser.parse_args()

    sthdata = train.load_data_with_pdata(args.patch_path)
    df = sthdata.adata.obs[["array_row", "array_col", "STHD_pred_ct"]]
    df_rasterize = rasterize(df, "STHD_pred_ct")
    fast_plot(df_rasterize, title=args.title, save_root_dir=args.patch_path)



import anndata
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


def simulate_spot_bin_2cell(
    side_length=15,
    num_points=16,
    center1=(4, 4),
    radius1=4,
    center2=(9, 9),
    radius2=4,
):
    """Simulator: input size in number of spot; define cell center and radius, output simulated spatial locations for 2 true cells, and bin boundary.

    Return:
    ------
    x,y. location of spot on x or y
    x_grid, y_grid. grid location of spot on x and y
    binxs, binys. bin boundary of 4 by 4 spot
    mask1, mask2. mask of cell1 and cell2 in x_grid and y_grid.

    Example:
    -------
    x,y,x_grid, y_grid, binxs, binys, mask1, mask2 = simulate_spot_bin_2cell(
        side_length = 15, num_points = 16,
        center1 = (4.5, 4.5) ,
        radius1 = 3,
        center2 = (9, 8) ,
        radius2 =4,
        )

    """
    ########## 2um spots ##########
    # Define the size of the square region
    # side_length = 15
    # Define the number of grid points along each axis
    # num_points = 16  # You can adjust this to change the density of the grid

    # Generate grid points within the square region
    # x = np.linspace(-side_length / 2, side_length / 2, num_points)
    x = np.linspace(0, side_length, num_points)
    y = np.linspace(0, side_length, num_points)
    # y = np.linspace(-side_length / 2, side_length / 2, num_points)
    x_grid, y_grid = np.meshgrid(x, y)

    # Plotting
    fig, ax = plt.subplots(figsize=(8, 8))
    ax.scatter(x_grid, y_grid, color="blue")
    plt.xlabel("X-axis")
    plt.ylabel("Y-axis")
    # plt.xlim(-side_length / 2-1, side_length / 2+1)
    # plt.xlim(0-1, side_length+1)
    # plt.ylim(-side_length / 2-1, side_length / 2+1)
    # plt.ylim(0-1, side_length+1)
    # plt.gca().set_aspect('equal', adjustable='box')
    # plt.grid(False)

    ########## 8um bins ##########
    ## add bin boundary
    lst = x + 0.5
    N = 4
    binxs = lst[N - 1 :: N]
    for xt in binxs:
        ax.axvline(xt, color="gray")

    lst = y + 0.5
    binys = lst[N - 1 :: N]
    N = 4
    for yt in binys:
        ax.axhline(yt, color="gray")

    ######## cell mask ##########

    # center1 = (4, 4)
    # radius1 = 4
    mask1 = ((x_grid - center1[0]) ** 2 + (y_grid - center1[1]) ** 2) <= radius1**2

    # center2 = (9, 9)
    # radius2 =4
    mask2 = ((x_grid - center2[0]) ** 2 + (y_grid - center2[1]) ** 2) <= radius2**2
    # remove mask1 in mask2, to make sure per spot is from one cell.
    mask_overlap = mask1 & mask2
    mask1[mask_overlap] = False

    ax.scatter(x_grid[mask1], y_grid[mask1], color="orange")
    ax.scatter(x_grid[mask2], y_grid[mask2], color="green")
    plt.show()
    return (x, y, x_grid, y_grid, binxs, binys, mask1, mask2)


def simulate_spot_expr_2cell(
    x_grid,
    y_grid,
    mask1,
    mask2,
    lam_ct1geneA=2,
    lam_ct1geneB=0,
    lam_ct2geneA=0.01,
    lam_ct2geneB=4,
    lam_geneC=4,
    lam_noise=0.001,
):
    """Example:
    -------
    adata = simulate_spot_expr_2cell(
        x_grid, y_grid, mask1, mask2,
        lam_ct1geneA=10,
        lam_ct1geneB=0,
        lam_ct2geneA=0.01,
        lam_ct2geneB=4,
        lam_noise=0.001,
        )

    """
    # create expression of spots for 2 cells of 2 cell types, marked by 2 marker genes.

    ## create an empty expr adata
    total_spot_num = x_grid.shape[0] * x_grid.shape[1]
    spotid_lst = [t for t in range(total_spot_num)]
    barcode_lst = ["barcode_" + str(t) for t in range(total_spot_num)]

    obs = pd.DataFrame(
        {"cellid": spotid_lst},
        index=barcode_lst,
    )
    var = pd.DataFrame(index=["geneA", "geneB", "geneC"])
    expr = pd.DataFrame(np.zeros([obs.shape[0], var.shape[0]]))
    expr.index = obs.index
    expr.columns = var.index

    # add true cell info
    iscell1_i = np.array(spotid_lst)[(mask1).flatten()]
    iscell2_i = np.array(spotid_lst)[(mask2).flatten()]

    obs["celltype"] = ""
    obs["iscell"] = False
    obs.loc[obs.iloc[iscell1_i].index, "celltype"] = "ct1"
    obs.loc[obs.iloc[iscell2_i].index, "celltype"] = "ct2"
    obs.loc[obs.iloc[iscell1_i].index, "iscell"] = True

    ## location of each spot
    obs["x"] = x_grid.flatten()
    obs["y"] = y_grid.flatten()
    obsm = {"spatial": obs[["x", "y"]].values}

    # simulate expression
    ct1_ids = obs[obs["celltype"] == "ct1"].index
    ct2_ids = obs[obs["celltype"] == "ct2"].index

    ct1geneA = np.random.poisson(lam=lam_ct1geneA, size=[len(ct1_ids)])
    ct1geneB = np.random.poisson(lam=lam_ct1geneB, size=[len(ct1_ids)])
    ct2geneA = np.random.poisson(lam=lam_ct2geneA, size=[len(ct2_ids)])
    ct2geneB = np.random.poisson(lam=lam_ct2geneB, size=[len(ct2_ids)])
    # print(ct1geneA)
    # print( expr.loc[ct1_ids]['geneA'].values)
    expr.loc[ct1_ids, "geneA"] = expr.loc[ct1_ids]["geneA"].values + ct1geneA
    # print( expr.loc[ct1_ids]['geneA'].values)
    expr.loc[ct1_ids, "geneB"] = expr.loc[ct1_ids]["geneB"].values + ct1geneB
    expr.loc[ct2_ids, "geneA"] = expr.loc[ct2_ids]["geneA"].values + ct2geneA
    expr.loc[ct2_ids, "geneB"] = expr.loc[ct2_ids]["geneB"].values + ct2geneB

    # simulate background noise
    geneN = np.random.poisson(lam=lam_noise, size=obs.shape[0])
    geneC = np.random.poisson(lam=lam_geneC, size=obs.shape[0])
    expr["geneA"] = expr["geneA"].values + geneN
    expr["geneB"] = expr["geneB"].values + geneN
    # add a 3rd unrelated gene
    expr.loc[:, "geneC"] = geneC

    adata = anndata.AnnData(expr, obs=obs, var=var, obsm=obsm)

    return adata



from collections import Counter

import numpy as np
from numba import njit, prange
from tqdm import tqdm

from STHD import model

## also try: T cell spot distance to myeloid


def get_neighbor_ct(adata, ctstr="cTNI", ctlst=[]):
    subset_cell_idx1 = np.where(adata.obs["STHD_pred_ct"].str.contains(ctstr))[
        0
    ]  # T cell id
    subset_cell_idx2 = np.where(adata.obs["STHD_pred_ct"].isin(ctlst))[0]
    subset_cell_idx = np.array(list(set(np.append(subset_cell_idx1, subset_cell_idx2))))
    _, idx = adata.obsp["spatial_connectivities"][
        subset_cell_idx.T, :
    ].nonzero()  # T cell neighbor id

    idx = np.array(
        list(set(np.append(idx, subset_cell_idx)))
    )  # T cell neighbor id plus T cell id
    subset_neighbor = adata.obs.iloc[idx]
    return subset_neighbor


def get_ambiguous_near_ct(adata, ctstr="cTNI", ctlst=[]):
    """Grep all ambiguous for an input neighbor identity. require sthdata.adata.obsp["spatial_connectivities"] from:
    # sq.gr.spatial_neighbors(sthdata.adata, n_rings=1, coord_type="grid", n_neighs=4)
    """
    #    sthdata = adata.copy()
    subset_cell_idx1 = np.where(adata.obs["STHD_pred_ct"].str.contains(ctstr))[
        0
    ]  # T cell id
    subset_cell_idx2 = np.where(adata.obs["STHD_pred_ct"].isin(ctlst))[0]
    subset_cell_idx = np.array(list(set(np.append(subset_cell_idx1, subset_cell_idx2))))
    _, idx = adata.obsp["spatial_connectivities"][
        subset_cell_idx.T, :
    ].nonzero()  # T cell neighbor id

    idx = np.array(
        list(set(np.append(idx, subset_cell_idx)))
    )  # T cell neighbor id plus T cell id

    # subset_neighbor_cell_idx = np.where(sthdata.adata.obs['STHD_pred_ct'].iloc[idx,'STHD_pred_ct']=='ambiguous')[0] # among neighbor, which one is the ambiguous cell type - those id
    # idx = np.append(idx, subset_cell_idx)
    amb_near_subset = adata.obs.iloc[idx][
        adata.obs["STHD_pred_ct"].iloc[idx] == "ambiguous"
    ].index  # id of ambiguous in T cell neighbor

    # ambiguous spots nearby T cell cTNI cells
    subset_neighbor = adata[amb_near_subset].copy()
    return subset_neighbor


######## frontline 1,3, 2,4 ############
def sthd_neighbor_ct_count(adata):
    # STHD probability columns
    sthd_p_cols = [t for t in adata.obs.columns if t[:5] == "p_ct_"]

    # obtain adjacency indices
    adj = adata.obsp["spatial_connectivities"]
    adj_row = adj.indptr
    adj_col = adj.indices

    # extract STHD probability, and use the max one to assign a celltype.
    p = adata.obs[sthd_p_cols]
    cell_type_names = list(p.columns)
    pmap = np.array([cell_type_names[i] for i in p.values.argmax(1)])
    pmap_value = p.values.max(1)
    print("[Log] remove filtered")
    pmap[pmap_value == -1] = "filtered"

    # count neighbour cell types
    neighbors = [
        model.csr_obtain_column_index_for_row(adj_row, adj_col, i)
        for i in range(len(p))
    ]
    neighbor_celltypes = [Counter(pmap[n]) for n in tqdm(neighbors)]
    neighbor_celltype_count = np.array([len(n_ct) for n_ct in tqdm(neighbor_celltypes)])
    print("[Log] Counting neighbor diversity...")
    print(Counter(neighbor_celltype_count))
    # give a name
    neighbor_celltype_name = np.array(
        ["|".join(sorted(list(n_ct.keys()))) for n_ct in tqdm(neighbor_celltypes)]
    )

    # put to data
    adata.obs["neighbor_celltype_name"] = neighbor_celltype_name
    adata.obs["neighbor_celltype_count"] = neighbor_celltype_count
    adata.obs["STHD_pred_ct_raw"] = pmap

    print(
        "[Log] neighbor information in adata.obs columns: neighbor_celltype_name, neighbor_celltype_count, STHD_pred_ct_raw"
    )


def sthd_neighbor_ct_count_binned(adata):
    # obtain adjacency indices
    adj = adata.obsp["spatial_connectivities"]
    adj_row = adj.indptr
    adj_col = adj.indices

    # count neighbour cell types
    neighbors = [
        model.csr_obtain_column_index_for_row(adj_row, adj_col, i)
        for i in range(len(adata))
    ]
    neighbor_celltypes = [
        Counter(adata.obs["STHD_pred_ct"].values[n]) for n in tqdm(neighbors)
    ]
    neighbor_celltype_count = np.array([len(n_ct) for n_ct in tqdm(neighbor_celltypes)])
    print("[Log] Counting neighbor diversity...")
    print(Counter(neighbor_celltype_count))
    # give a name
    neighbor_celltype_name = np.array(
        ["|".join(sorted(list(n_ct.keys()))) for n_ct in tqdm(neighbor_celltypes)]
    )

    # put to data
    adata.obs["neighbor_celltype_name"] = neighbor_celltype_name
    adata.obs["neighbor_celltype_count"] = neighbor_celltype_count
    print(
        "[Log] neighbor information in adata.obs columns: neighbor_celltype_name, neighbor_celltype_count, STHD_pred_ct_raw"
    )


def get_frontline(
    adata, A="Tumor cE", B="Macrophage", frontline_name="frontline_ctA_ctB"
):
    # given cell types A,B (string containing pattern),
    itself_A_B = (adata.obs["STHD_pred_ct"].str.contains(A).values) | (
        adata.obs["STHD_pred_ct"].str.contains(B).values
    )
    subset = adata[(itself_A_B) & (adata.obs["neighbor_celltype_count"] <= 2)].copy()
    neighbor_A_B = (subset.obs["neighbor_celltype_name"].str.contains(A).values) & (
        subset.obs["neighbor_celltype_name"].str.contains(B).values
    )
    label = np.zeros(len(subset))  # initialize as all zeros

    # A
    label[
        (subset.obs["STHD_pred_ct"].str.contains(A).values)
        & (subset.obs["neighbor_celltype_count"] == 1)
    ] = 1
    # B
    label[
        (subset.obs["STHD_pred_ct"].str.contains(B).values)
        & (subset.obs["neighbor_celltype_count"] == 1)
    ] = 2
    # xA
    label[(subset.obs["STHD_pred_ct"].str.contains(A).values) & neighbor_A_B] = 3
    # xB
    label[(subset.obs["STHD_pred_ct"].str.contains(B).values) & neighbor_A_B] = 4
    print("[Log] Counting frontline spot... ")
    print(Counter(label))

    subset.obs["label"] = label

    adata.obs[frontline_name] = 0
    labels = adata.obs[frontline_name].values
    labels[
        np.where(adata.obs.index.isin(subset[subset.obs["label"] == 1].obs.index))[0]
    ] = 1
    labels[
        np.where(adata.obs.index.isin(subset[subset.obs["label"] == 2].obs.index))[0]
    ] = 2
    labels[
        np.where(adata.obs.index.isin(subset[subset.obs["label"] == 3].obs.index))[0]
    ] = 3
    labels[
        np.where(adata.obs.index.isin(subset[subset.obs["label"] == 4].obs.index))[0]
    ] = 4
    adata.obs[frontline_name] = labels
    print(
        f'frontline types in adata.obs["{frontline_name}"]',
        set(adata.obs[frontline_name]),
    )


def get_frontline_binned(
    adata, A="Tumor cE", B="Macrophage", frontline_name="frontline_ctA_ctB"
):
    # given cell types A,B (string containing pattern),
    itself_A_B = (adata.obs["STHD_pred_ct"].str.contains(A).values) | (
        adata.obs["STHD_pred_ct"].str.contains(B).values
    )
    subset = adata[(itself_A_B)].copy()
    neighbor_A_B = (subset.obs["neighbor_celltype_name"].str.contains(A).values) & (
        subset.obs["neighbor_celltype_name"].str.contains(B).values
    )
    label = np.zeros(len(subset))  # initialize as all zeros

    # A
    label[
        (subset.obs["STHD_pred_ct"].str.contains(A).values)
        & (subset.obs["neighbor_celltype_count"] == 1)
    ] = 1
    # B
    label[
        (subset.obs["STHD_pred_ct"].str.contains(B).values)
        & (subset.obs["neighbor_celltype_count"] == 1)
    ] = 2
    # xA
    label[(subset.obs["STHD_pred_ct"].str.contains(A).values) & neighbor_A_B] = 3
    # xB
    label[(subset.obs["STHD_pred_ct"].str.contains(B).values) & neighbor_A_B] = 4
    print("[Log] Counting frontline spot... ")
    print(Counter(label))

    subset.obs["label"] = label

    adata.obs[frontline_name] = 0
    labels = adata.obs[frontline_name].values
    labels[
        np.where(adata.obs.index.isin(subset[subset.obs["label"] == 1].obs.index))[0]
    ] = 1
    labels[
        np.where(adata.obs.index.isin(subset[subset.obs["label"] == 2].obs.index))[0]
    ] = 2
    labels[
        np.where(adata.obs.index.isin(subset[subset.obs["label"] == 3].obs.index))[0]
    ] = 3
    labels[
        np.where(adata.obs.index.isin(subset[subset.obs["label"] == 4].obs.index))[0]
    ] = 4
    adata.obs[frontline_name] = labels
    print(
        f'frontline types in adata.obs["{frontline_name}"]',
        set(adata.obs[frontline_name]),
    )


def frontline_summarize(adata, frontlines):
    # summarize the frontline type.
    adata.obs["frontline_sum_type"] = "non_fl"
    labels = adata.obs["frontline_sum_type"].values
    for frontline in frontlines:
        labels[adata.obs[frontline].isin([3, 4])] = frontline
    adata.obs["frontline_sum_type"] = labels

    adata.obs["frontline_sum_type_AB"] = "non_fl"
    labels = adata.obs["frontline_sum_type_AB"].values
    for frontline in frontlines:
        labels[adata.obs[frontline] == 3] = frontline + "_A"
        labels[adata.obs[frontline] == 4] = frontline + "_B"
    adata.obs["frontline_sum_type_AB"] = labels
    print(
        "[Log] Summarized frontline type in frontline_sum_type, and frontline_sum_type_AB"
    )


######## Distance to front lines ############
@njit(parallel=True, fastmath=True)
def min_pairwise_distance(X, Y):
    # X and Y represents 2 sets of 2D dots.
    # For each dot in X, find the cloest dot in Y and calculate their euclidian distance
    n, m = len(X), len(Y)
    res = np.zeros(n, dtype=np.float32)
    for i in prange(n):
        x = X[i]
        d = np.inf
        for j in range(m):
            y = Y[j]
            cur_d = ((x[0] - y[0]) ** 2 + (x[1] - y[1]) ** 2) ** (0.5)
            d = min(d, cur_d)
        res[i] = d
    return res


def calculate_distance(adata, frontline_name):
    # assuming, label of 3 and 4 indicates xA and xB, which is the frontline.
    frontline_label = np.isin(adata.obs[frontline_name].values, [3, 4])
    # location = adata.obs[['x','y']].values
    location = adata.obs[["array_row", "array_col"]].values
    frontline_location = location[frontline_label]
    distance = min_pairwise_distance(location, frontline_location)
    adata.obs["dTo_" + frontline_name] = distance
    print(
        f'Distance to front line {frontline_name} is saved in adata.obs["dTo_{frontline_name}"]'
    )


def calculate_distance_binned(adata, frontline_name):
    # assuming, label of 3 and 4 indicates xA and xB, which is the frontline.
    frontline_label = np.isin(adata.obs[frontline_name].values, [3, 4])
    # location = adata.obs[['x','y']].values
    location = adata.obs[["bin_row", "bin_col"]].values.astype(int)
    frontline_location = location[frontline_label]
    distance = min_pairwise_distance(location, frontline_location)
    adata.obs["dTo_" + frontline_name] = distance
    print(
        f'Distance to front line {frontline_name} is saved in adata.obs["dTo_{frontline_name}"]'
    )



## low neighbor count detection to mask some spots with local continuous low count in neighbors.
import matplotlib.pyplot as plt
import numpy as np
import squidpy as sq
from numba import njit


def background_detector(
    adata, layer=None, threshold=50, n_neighs=4, n_rings=4, coord_type="grid"
):
    """Compute neighbor using spatial, default n_neighs =4, n_rings=2
    Add columns in adata.obs: 'neighbor_agg_counts', 'neighbor_agg_log_counts', 'low_count_region'
    Params
    ----------
    adata: scanpy obj with spatial coords
    threshold
        low count filter thresold. e.g. 50 cutoff for aggregated neighbor counts
    layer
        layer in adata used for neighbor counter filtering. should be the layer with raw counts. e.g. layer='count', defaul: adata.X
    n_rings = 2
        # number of rings to check on neighbors
    n_neighs
        For square shape, n_neighs=4. and coord_type = 'grid'
    coord_type
        For square shape, n_neighs=4. and coord_type = 'grid'

    """
    if layer == "count":
        adataX = adata.layers["count"]
    else:
        adataX = adata.X
    X = adata.obs.shape[0]  # n of spot
    D_raw = np.squeeze(
        np.asarray(adataX.sum(axis=1))
    )  # [X] per spot counts. IMPORTANT: DO NOT use .to_df - slow
    # print(D_raw[:10])
    p_background = D_raw.copy()
    sq.gr.spatial_neighbors(
        adata,
        spatial_key="spatial",
        coord_type=coord_type,
        n_neighs=n_neighs,
        n_rings=n_rings,
    )
    A_csr = adata.obsp["spatial_distances"]  # [X, X]
    Acsr_row = A_csr.indptr
    Acsr_col = A_csr.indices
    Acsr_val = A_csr.data
    aggregate_neighbor(D_raw, p_background, Acsr_row, Acsr_col, Acsr_val, X)
    adata.obs["neighbor_agg_counts"] = p_background
    adata.obs["neighbor_agg_log_counts"] = np.log(p_background + 1)
    adata.obs["low_count_region"] = (p_background < threshold).astype(int)
    return adata


def visualize_background(sthdata):
    x1, y1, x2, y2 = sthdata.get_sequencing_data_region()
    f, ax = plt.subplots(2, 2, figsize=(15, 15))
    # histogram
    _ = ax[0, 0].hist(
        sthdata.adata.obs["neighbor_agg_counts"],
        bins=500,
    )
    ax[0, 0].set_xlim([-10, 300])
    ax[0, 0].set_title("aggregated neighbor coutns")
    # neighbor agg log counts
    sq.pl.spatial_scatter(
        sthdata.adata,
        color=["neighbor_agg_log_counts"],
        crop_coord=[(x1, y1, x2, y2)],
        # shape='square',
        legend_fontsize=6,
        ax=ax[0, 1],
    )
    #
    img = sthdata.load_img()
    img_crop = sthdata.crop_img(img, x1, y1, x2, y2)
    ax[1, 0].imshow(img_crop)
    # islowcount
    sq.pl.spatial_scatter(
        sthdata.adata,
        color=["low_count_region"],
        crop_coord=[(x1, y1, x2, y2)],
        legend_fontsize=6,
        ax=ax[1, 1],
    )


def filter_background(
    sthdata,
    threshold=50,
    n_neighs=4,
    n_rings=4,
    coord_type="grid",
    low_cts_col="low_count_region",
    inplace=False,
):
    """Filter out low neighbor count barcodes, and return filtered sthd
    Params
    ----------
    sthdata
        STHD obj
    threshold
        low count filter thresold. e.g. 100 cutoff for aggregated neighbor counts
    n_rings = 2
        # number of rings to check on neighbors
    n_neighs
        For square shape, n_neighs=4. and coord_type = 'grid'
    coord_type
        For square shape, n_neighs=4. and coord_type = 'grid'
    low_cts_col
        column to record whether a barcode is in low count
    #neighbor_cts_col
    #    column to record neighbor count aggregated
    inplace
        default False, assign True to save memory for full slide data
    """
    if inplace:
        sthdata_filtered = sthdata
    else:
        sthdata_filtered = sthdata.copy()
    sthdata_filtered.adata = background_detector(
        sthdata_filtered.adata,
        threshold=threshold,
        n_neighs=n_neighs,
        n_rings=n_rings,
        coord_type=coord_type,
    )
    before = sthdata_filtered.adata.shape[0]
    sthdata_filtered.adata = sthdata_filtered.adata[
        sthdata_filtered.adata.obs[low_cts_col] < 1
    ]
    after = sthdata_filtered.adata.shape[0]
    print(f"[Log] filtering background: {before} spots to {after} spots")
    return sthdata_filtered


@njit
def csr_obtain_column_val_for_row(row, column, val, i):
    # get row i's non-zero items' column index
    row_start = row[i]
    row_end = row[i + 1]
    column_indices = column[row_start:row_end]
    values = val[row_start:row_end]
    return column_indices, values


@njit
def aggregate_neighbor(D_raw, p_background, row, column, val, X):
    for a in range(X):
        neighbors, distances = csr_obtain_column_val_for_row(row, column, val, a)
        for i in range(len(neighbors)):
            neighbor = neighbors[i]
            distance = distances[i]
            p_background[a] += D_raw[neighbor] / (distance + 1)
